//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/viewer.h"
#include "TransProg/ptrVector.h"
#include "TransProg/random.h"
#include "TransProg/text.h"
#include "TransProg/path.h"
#include "TransGl/glDraw.h"

const real32_t tetraVertices[4*3]={  1.0,  1.0,  1.0,
                                    -1.0, -1.0,  1.0,
                                    -1.0,  1.0, -1.0,
                                     1.0, -1.0, -1.0 };
const real32_t tetraColors[4*3]={ 1.0, 0.0, 0.0,
                                  0.0, 1.0, 0.0,
                                  0.0, 0.0, 1.0,
                                  1.0, 1.0, 1.0 };
const uint8_t tetraIndices[4*3]={ 0, 2, 1,
                                  1, 2, 3,
                                  2, 0, 3,
                                  3, 0, 1 };

const real32_t boxVertices[24*3]={ -1.0, -1.0, -1.0, // Z-
                                   -1.0,  1.0, -1.0,
                                    1.0,  1.0, -1.0,
                                    1.0, -1.0, -1.0,
                                   -1.0, -1.0, -1.0, // Y-
                                    1.0, -1.0, -1.0,
                                    1.0, -1.0,  1.0,
                                   -1.0, -1.0,  1.0,
                                    1.0, -1.0, -1.0, // X+
                                    1.0,  1.0, -1.0,
                                    1.0,  1.0,  1.0,
                                    1.0, -1.0,  1.0,
                                    1.0,  1.0, -1.0, // Y+
                                   -1.0,  1.0, -1.0,
                                   -1.0,  1.0,  1.0,
                                    1.0,  1.0,  1.0,
                                   -1.0,  1.0, -1.0, // X-
                                   -1.0, -1.0, -1.0,
                                   -1.0, -1.0,  1.0,
                                   -1.0,  1.0,  1.0,
                                   -1.0, -1.0,  1.0, // Z+
                                    1.0, -1.0,  1.0,
                                    1.0,  1.0,  1.0,
                                   -1.0,  1.0,  1.0 };
const real32_t boxNormals[24*3]={  0.0,  0.0, -1.0,
                                   0.0,  0.0, -1.0,
                                   0.0,  0.0, -1.0,
                                   0.0,  0.0, -1.0,
                                   0.0, -1.0,  0.0,
                                   0.0, -1.0,  0.0,
                                   0.0, -1.0,  0.0,
                                   0.0, -1.0,  0.0,
                                   1.0,  0.0,  0.0,
                                   1.0,  0.0,  0.0,
                                   1.0,  0.0,  0.0,
                                   1.0,  0.0,  0.0,
                                   0.0,  1.0,  0.0,
                                   0.0,  1.0,  0.0,
                                   0.0,  1.0,  0.0,
                                   0.0,  1.0,  0.0,
                                  -1.0,  0.0,  0.0,
                                  -1.0,  0.0,  0.0,
                                  -1.0,  0.0,  0.0,
                                  -1.0,  0.0,  0.0,
                                   0.0,  0.0,  1.0,
                                   0.0,  0.0,  1.0,
                                   0.0,  0.0,  1.0,
                                   0.0,  0.0,  1.0 };
const real32_t boxColors[24*3]={ 1.0, 0.0, 0.0,
                                 1.0, 0.0, 0.0,
                                 1.0, 0.0, 0.0,
                                 1.0, 0.0, 0.0,
                                 0.0, 1.0, 0.0,
                                 0.0, 1.0, 0.0,
                                 0.0, 1.0, 0.0,
                                 0.0, 1.0, 0.0,
                                 0.0, 0.0, 1.0,
                                 0.0, 0.0, 1.0,
                                 0.0, 0.0, 1.0,
                                 0.0, 0.0, 1.0,
                                 1.0, 1.0, 0.0,
                                 1.0, 1.0, 0.0,
                                 1.0, 1.0, 0.0,
                                 1.0, 1.0, 0.0,
                                 1.0, 0.0, 1.0,
                                 1.0, 0.0, 1.0,
                                 1.0, 0.0, 1.0,
                                 1.0, 0.0, 1.0,
                                 0.0, 1.0, 1.0,
                                 0.0, 1.0, 1.0,
                                 0.0, 1.0, 1.0,
                                 0.0, 1.0, 1.0 };
const uint8_t boxIndices[12*3]={ 0,  1,  2,  0,  2,  3,
                                 4,  5,  6,  4,  6,  7,
                                 8,  9, 10,  8, 10, 11,
                                12, 13, 14, 12, 14, 15,
                                16, 17, 18, 16, 18, 19,
                                20, 21, 22, 20, 22, 23 };

#define NB_POINTS 15

typedef struct _Object
{
GlShader shader;
int v_Vertex,v_Color,v_Normal;
int u_ProjectionMatrix,u_ModelViewMatrix;
GlVertexArray vertexArray;
char kind;
SpatialBasis location;
} Object;

Viewer *viewer;
PtrVector objects;

void
viewerCB(Viewer *v)
{
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
Viewer_setRefDistanceFromPoint(v,0.0,0.0,0.0);
bool close=false;
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"\x1B"))
        {
        close=true;
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    real32_t modelViewMtx[16];
    for(intptr_t i=0,i_end=PtrVector_size(&objects);i<i_end;++i)
      {
      Object *o=(Object *)PtrVector_get(&objects,i);
      Gfx_computeModelViewMatrix(modelViewMtx,viewMtx,o->location.matrix);
      glUseProgram(GlShader_program(&o->shader));
      glUniformMatrix4fv(o->u_ProjectionMatrix,1,GL_FALSE,projMtx);
      glUniformMatrix4fv(o->u_ModelViewMatrix,1,GL_FALSE,modelViewMtx);
      GlVertexArray_bind(&o->vertexArray);
      switch(o->kind)
        {
        case 'A':
        case 'B':
          {
          glDrawRangeElements(GL_TRIANGLES,0,3,
                              4*3,GL_UNSIGNED_BYTE,(const void *)0);
          break;
          }
        case 'C':
          {
          glDrawRangeElements(GL_TRIANGLES,0,23,
                              12*3,GL_UNSIGNED_BYTE,(const void *)0);
          break;
          }
        case 'D':
          {
          glDrawArrays(GL_POINTS,0,NB_POINTS);
          break;
          }
        }
      }
    break;
    }
  default:
    {
    break;
    }
  }
if(close)
  {
  Viewer_delete(v);
  viewer=(Viewer *)0;
  }
}

Text *
loadFile(const char *fileName)
{
intptr_t sz=(intptr_t)Path_fileSize(fileName);
if(sz<0)
  {
  return (Text *)0;
  }
Text *t=Text_new();
Text_ensureMinCapacity(t,(intptr_t)sz);
char *s=Text_get(t);
if(Path_readFile(fileName,s,sz,0)!=sz)
  {
  Text_delete(t);
  return (Text *)0;
  }
s[sz]='\0';
t->v.size=sz;
return t;
}

Object *
makeObject(char kind)
{
int cpuCtx=Platform_cpuCtx();
Object *o=Memory_alloc(1,Object);
Text *src[3]={(Text *)0,(Text *)0,(Text *)0};
switch(kind)
  {
  case 'A':
    {
    src[0]=loadFile("sh_A.vert");
    src[2]=loadFile("sh_default.frag");
    break;
    }
  case 'B':
    {
    src[0]=loadFile("sh_B.vert");
    src[1]=loadFile("sh_B.geom");
    src[2]=loadFile("sh_default.frag");
    break;
    }
  case 'C':
    {
    src[0]=loadFile("sh_C.vert");
    src[2]=loadFile("sh_default.frag");
    break;
    }
  case 'D':
    {
    src[0]=loadFile("sh_D.vert");
    src[1]=loadFile("sh_D.geom");
    src[2]=loadFile("sh_default.frag");
    break;
    }
  default:
    {
    DROP_DEAD("object kind?");
    }
  }
GlShader_init(&o->shader,
              src[0] ? Text_get(src[0]) : (const char *)0,
              src[1] ? Text_get(src[1]) : (const char *)0,
              src[2] ? Text_get(src[2]) : (const char *)0,
              true);
int program=GlShader_program(&o->shader);
if(!program)
  {
  const char *log=GlShader_log(&o->shader);
  if(log)
    {
    Platform_msg("~~~~~~~~~~~~~~~~ %c ~~~~~~~~~~~~~~~~\n%s\n",kind,log);
    }
  DROP_DEAD("shader build failure");
  }
glUseProgram(program);
o->v_Vertex=glGetAttribLocation(program,"v_Vertex");
o->v_Color=glGetAttribLocation(program,"v_Color");
if(kind=='C')
  {
  o->v_Normal=glGetAttribLocation(program,"v_Normal");
  }
o->u_ProjectionMatrix=glGetUniformLocation(program,"u_ProjectionMatrix");
o->u_ModelViewMatrix=glGetUniformLocation(program,"u_ModelViewMatrix");
GlVertexArray_init(&o->vertexArray);
switch(kind)
  {
  case 'A':
  case 'B':
    {
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Vertex,3,0,
                               4,tetraVertices);
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Color,3,0,
                               4,tetraColors);
    GlVertexArray_createBuffer(&o->vertexArray,
                               -1,-1,0,
                               4*3,tetraIndices);
    break;
    }
  case 'C':
    {
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Vertex,3,0,
                               24,boxVertices);
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Color,3,0,
                               24,boxColors);
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Normal,3,0,
                               24,boxNormals);
    GlVertexArray_createBuffer(&o->vertexArray,
                               -1,-1,0,
                               12*3,boxIndices);
    break;
    }
  case 'D':
    {
    real32_t point[3*NB_POINTS];
    real32_t pointColor[3*NB_POINTS];
    for(int i=0;i<3*NB_POINTS;i+=3)
      {
      point[i+0]=2.0f*(Random_real32(cpuCtx)-0.5f);
      point[i+1]=2.0f*(Random_real32(cpuCtx)-0.5f);
      point[i+2]=2.0f*(Random_real32(cpuCtx)-0.5f);
      pointColor[i+0]=Random_real32(cpuCtx);
      pointColor[i+1]=Random_real32(cpuCtx);
      pointColor[i+2]=Random_real32(cpuCtx);
      }
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Vertex,3,0,
                               NB_POINTS,point);
    GlVertexArray_createBuffer(&o->vertexArray,
                               o->v_Color,3,0,
                               NB_POINTS,pointColor);
    break;
    }
  default:
    {
    DROP_DEAD("object kind ?");
    }
  }
o->kind=kind;
SpatialBasis_init(&o->location);
SpatialBasis_translate(&o->location,
                       10.0*(Random_real64(cpuCtx)-0.5),
                       10.0*(Random_real64(cpuCtx)-0.5),
                       10.0*(Random_real64(cpuCtx)-0.5));
return o;
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
GlView_init();
GlDraw_init();
Platform_msg("OpenGL version: %s%d.%d%s\n",
             CoreGl_embedded() ? "ES " : "",
             CoreGl_majorVersion(),CoreGl_minorVersion(),
             CoreGl_coreProfile() ? " core profile" : "");
Platform_msg("GLSL version: %d\n",CoreGl_shadingLanguageVersion());
int extCount;
const char * const *ext=CoreGl_extensions(&extCount);
for(int i=0;i<extCount;++i)
  {
  Platform_msg("%s\n",ext[i]);
  }
if(CoreGl_has_VERSION_1_0())
  Platform_msg("VERSION_1_0\n");
if(CoreGl_has_VERSION_1_1())
  Platform_msg("VERSION_1_1\n");
if(CoreGl_has_VERSION_1_2())
  Platform_msg("VERSION_1_2\n");
if(CoreGl_has_VERSION_1_3())
  Platform_msg("VERSION_1_3\n");
if(CoreGl_has_VERSION_1_4())
  Platform_msg("VERSION_1_4\n");
if(CoreGl_has_VERSION_1_5())
  Platform_msg("VERSION_1_5\n");
if(CoreGl_has_VERSION_2_0())
  Platform_msg("VERSION_2_0\n");
if(CoreGl_has_VERSION_2_1())
  Platform_msg("VERSION_2_1\n");
if(CoreGl_has_VERSION_3_0())
  Platform_msg("VERSION_3_0\n");
if(CoreGl_has_VERSION_3_1())
  Platform_msg("VERSION_3_1\n");
if(CoreGl_has_VERSION_3_2())
  Platform_msg("VERSION_3_2\n");
if(CoreGl_has_VERSION_3_3())
  Platform_msg("VERSION_3_3\n");
if(CoreGl_has_VERSION_4_0())
  Platform_msg("VERSION_4_0\n");
if(CoreGl_has_VERSION_4_1())
  Platform_msg("VERSION_4_1\n");
if(CoreGl_has_VERSION_4_2())
  Platform_msg("VERSION_4_2\n");
if(CoreGl_has_VERSION_4_3())
  Platform_msg("VERSION_4_3\n");
if(CoreGl_has_VERSION_4_4())
  Platform_msg("VERSION_4_4\n");
if(CoreGl_has_VERSION_4_5())
  Platform_msg("VERSION_4_5\n");
Platform_msg("has tessellation control shaders: %s\n",
             CoreGl_hasTessControlShaders() ? "yes" : "no");
Platform_msg("has tessellation evaluation shaders: %s\n",
             CoreGl_hasTessEvaluationShaders() ? "yes" : "no");
Platform_msg("has geometry shaders: %s\n",
             CoreGl_hasGeometryShaders() ? "yes" : "no");
Platform_msg("has compute shaders: %s\n",
             CoreGl_hasComputeShaders() ? "yes" : "no");

int w=512;
int h=512;
viewer=Viewer_new(GlView_screenWidth()-w,0,w,h,false,&viewerCB,(void *)0);
Perspective_setNearDistance(Viewer_perspective(viewer),0.1);
Perspective_setFarDistance(Viewer_perspective(viewer),100.0);
SpatialBasis_translate(Viewer_viewPoint(viewer),-15.0,0.0,0.0);

PtrVector_init(&objects);
if(argc>1)
  {
  for(const char *c=argv[1];*c;++c)
    {
    PtrVector_push(&objects,makeObject(*c));
    }
  }
int64_t last=Platform_time();
for(;;)
  {
  int64_t t=Platform_time();
  real64_t dt=(t-last)*1e-6;
  last=t;
  GlView_processEvents(0);
  GlView_sync();
  if(!viewer) break;
  for(intptr_t i=0,i_end=PtrVector_size(&objects);i<i_end;++i)
    {
    Object *o=(Object *)PtrVector_get(&objects,i);
    SpatialBasis_roll(&o->location,-dt*0.1*M_PI);
    SpatialBasis_pitch(&o->location,dt*0.1*M_PI);
    SpatialBasis_yaw(&o->location,dt*0.05*M_PI);
    }
  GlView_postRedraw(Viewer_view(viewer));
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
