//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef PERF_H
#define PERF_H 1

#include "TransProg/memory.h"
#include "TransProg/random.h"

void
copy32(int count,
       const int * RESTRICT idx,
       const real32_t *in,
       real32_t *out);

void
copy64(int count,
       const int * RESTRICT idx,
       const real64_t *in,
       real64_t *out);

void
convIntTo32(int count,
            const int * RESTRICT idx,
            const int *in,
            real32_t *out);

void
conv32ToInt(int count,
            const int * RESTRICT idx,
            const real32_t *in,
            int *out);

void
convIntTo64(int count,
            const int * RESTRICT idx,
            const int *in,
            real64_t *out);

void
conv64ToInt(int count,
            const int * RESTRICT idx,
            const real64_t *in,
            int *out);

void
conv32To64(int count,
           const int * RESTRICT idx,
           const real32_t *in,
           real64_t *out);

void
conv64To32(int count,
           const int * RESTRICT idx,
           const real64_t *in,
           real32_t *out);

void
min32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out);

void
min64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out);

void
max32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out);

void
max64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out);

void
add32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out);

void
add64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out);

void
mul32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out);

void
mul64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out);

void
div32(int count,
      const int * RESTRICT idx,
      const real32_t *in,
      real32_t *out);

void
div64(int count,
      const int * RESTRICT idx,
      const real64_t *in,
      real64_t *out);

void
classic32(int count,
          const int * RESTRICT idx,
          const real32_t *in,
          real32_t *out);

void
approx32(int count,
         const int * RESTRICT idx,
         const real32_t *in,
         real32_t *out);

void
classic64(int count,
          const int * RESTRICT idx,
          const real64_t *in,
          real64_t *out);

void
approx64(int count,
         const int * RESTRICT idx,
         const real64_t *in,
         real64_t *out);

#endif // PERF_H

//----------------------------------------------------------------------------
