//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/iVector.h"

void
task0(int cpuCtx,
      void *data)
{
ASSERT_CURRENT_CPUCTX(cpuCtx);
(void)data;
for(int i=0;i<100000000;++i)
  {
  volatile char *p; // volatile avoids code removal
  p=(char *)malloc(1+(i%1000));
  p[0]='X';
  free((char *)p);
  }
}

void
task1(int cpuCtx,
      void *data)
{
ASSERT_CURRENT_CPUCTX(cpuCtx);
(void)data;
for(int i=0;i<100000000;++i)
  {
  volatile char *p; // volatile avoids code removal
  p=Memory_alloc(1+(i%1000),char);
  p[0]='X';
  Memory_free((char *)p);
  }
}

void
task2(int cpuCtx,
      void *data)
{
ASSERT_CURRENT_CPUCTX(cpuCtx);
(void)data;
for(int i=0;i<500;++i)
  {
  IVector *v=IVector_new();
  while(IVector_size(v)<10000)
    {
    IVector_push(v,(int)IVector_size(v));
    }
  IVector *v2=IVector_new();
  IVector_insertMany(v2,0,v->data,IVector_size(v));
  while(IVector_size(v2))
    {
    IVector_eraseMany(v2,0,2);
    }
  IVector_delete(v2);
  while(IVector_size(v))
    {
    IVector_erase(v,IVector_size(v)-1);
    }
  IVector_delete(v);
  }
}

void
task3(int cpuCtx,
      void *data)
{
ASSERT_CURRENT_CPUCTX(cpuCtx);
(void)data;
for(int i=0;i<10000;++i)
  {
  IVector *v1=IVector_new();
  IVector *v2=IVector_new();
  IVector *v3=IVector_new();
  while(IVector_size(v1)<10000)
    {
    IVector_push(v1,(int)IVector_size(v1));
    IVector_push(v2,(int)IVector_size(v2));
    IVector_push(v3,(int)IVector_size(v3));
    }
  IVector_delete(v1);
  IVector_delete(v2);
  IVector_delete(v3);
  }
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
void (*tasks[])(int,void *)={&task0,&task1,&task2,&task3};
int taskId=0;
if(argc>1)
  {
  sscanf(argv[1],"%d",&taskId);
  taskId%=sizeof(tasks)/sizeof(tasks[0]);
  }
Platform_msg("task %d\n",taskId);
int64_t dt=Platform_time();
Platform_parallel(tasks[taskId],(void *)0);
dt=Platform_time()-dt;
Platform_msg("duration: %"F64"d us\n",dt);
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
