//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_MEMORY_H
#define TRANSPROG_MEMORY_H 1

#include "TransProg/platform.h"

TRANSPROG_API
void * // reallocated memory
Memory_reAllocVoid(void *oldPtr,
                   intptr_t newSize);

#define Memory_reAlloc(oldPtr,newSize,type) \
        ((type *)Memory_reAllocVoid((oldPtr),(newSize)*sizeof(type)))

TRANSPROG_API
void * // allocated memory
Memory_allocVoid(intptr_t size);

#define Memory_alloc(size,type) \
        ((type *)Memory_allocVoid((size)*sizeof(type)))

TRANSPROG_INLINE
void * // allocated memory (0 initialized)
Memory_initAllocVoid(intptr_t size)
{
return memset(Memory_allocVoid(size),0,size);
}

#define Memory_initAlloc(size,type) \
        ((type *)Memory_initAllocVoid((size)*sizeof(type)))

TRANSPROG_INLINE
void * // allocated copy of data
Memory_allocCopyVoid(const void *data,
                     intptr_t size)
{
return memcpy(Memory_allocVoid(size),data,size);
}

#define Memory_allocCopy(data,size,type) \
        ((type *)Memory_allocCopyVoid(data,(size)*sizeof(type)))

TRANSPROG_INLINE
char * // allocated copy of str
Memory_allocStr(const char *str)
{
return Memory_allocCopy(str,strlen(str)+1,char);
}

TRANSPROG_API
void
Memory_free(void *ptr);

TRANSPROG_API
void * // allocated pages or 0
Memory_pageAlloc(intptr_t *inout_size,
                 bool preferHuge);

TRANSPROG_API
void
Memory_pageFree(void *pages,
                intptr_t size);

TRANSPROG_API
void
Memory_numaUsage(const void *ptr,
                 intptr_t *numaUsage);

TRANSPROG_INLINE
real64_t // numa context usage ratio in [0.0;1.0]
Memory_numaBalance(int numaCtx,
                   const intptr_t *numaUsage)
{
intptr_t usage=0;
for(int n=0,n_end=Platform_numaCtxCount();n<n_end;++n)
  {
  usage+=numaUsage[n];
  }
#if !defined NDEBUG
  if(!usage)
    {
    DROP_DEAD("no numa usage");
    }
#endif
return usage ? numaUsage[numaCtx]/(real64_t)usage : 0.0;
}

#endif // TRANSPROG_MEMORY_H

//----------------------------------------------------------------------------
