//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_PLATFORM_H
#define TRANSPROG_PLATFORM_H 1

#if defined __cplusplus
# define TRANSPROG_EXTERN_C extern "C"
#else
# define TRANSPROG_EXTERN_C
#endif

#if defined _WIN32
# define TRANSPROG_EXPORT_SYMBOL __declspec(dllexport)
# define TRANSPROG_IMPORT_SYMBOL __declspec(dllimport)
# if defined WIN32_TRANSPROG_DLL
#   define TRANSPROG_API_VAR TRANSPROG_EXPORT_SYMBOL
# else
#   define TRANSPROG_API_VAR TRANSPROG_IMPORT_SYMBOL
# endif
#else
# define TRANSPROG_EXPORT_SYMBOL
# define TRANSPROG_IMPORT_SYMBOL
# define TRANSPROG_API_VAR
#endif

#define TRANSPROG_API TRANSPROG_EXTERN_C TRANSPROG_API_VAR

#if defined _MSC_VER
# define TRANSPROG_INLINE __inline static
#else
# define TRANSPROG_INLINE inline static
#endif

#define RESTRICT __restrict__

#if defined __INTEL_COMPILER
# define TRANSPROG_VECTORIZE \
         _Pragma("ivdep") \
         _Pragma("vector always")
# define TRANSPROG_DONT_VECTORIZE \
         _Pragma("novector")
#elif defined _MSC_VER
# define TRANSPROG_VECTORIZE \
         __pragma(loop(ivdep))
# define TRANSPROG_DONT_VECTORIZE \
         __pragma(loop(no_vector))
#elif (__GNUC__*100+__GNUC_MINOR__)>=409
# define TRANSPROG_VECTORIZE \
         _Pragma("GCC ivdep")
# define TRANSPROG_DONT_VECTORIZE \
         // FIXME: nothing to prevent vectorization
#else
# define TRANSPROG_VECTORIZE
# define TRANSPROG_DONT_VECTORIZE
#endif

#if defined _WIN32
# define WINVER 0x601
# define _WIN32_WINNT WINVER
# if defined _MSC_VER
#   if !defined _DLL
#     error "_DLL undefined, use /MD when compiling!"
#   endif
#   if !defined _MT
#     error "_MT undefined, use /MD when compiling!"
#   endif
#   if !defined _DEBUG
#     define NDEBUG
#   endif
#   define BIG_ENDIAN    4321
#   define LITTLE_ENDIAN 1234
#   if defined _M_IX86 || defined _M_AMD64
#     define BYTE_ORDER LITTLE_ENDIAN // UGLY hard-coded setting!
#   endif
#   define _CRT_SECURE_NO_WARNINGS
#   include <stdint.h>
#   include <limits.h>
    typedef float real32_t;
    typedef double real64_t;
#   define __restrict__ __restrict
#   define __func__ __FUNCTION__
//  !!! thread is not compatible with dllexport
//  # define TRANSPROG_THREAD_STORAGE __declspec(thread)
#   define strcasecmp _stricmp
#   define strncasecmp _strnicmp
#   define snprintf _snprintf
# else
#   define __USE_MINGW_ANSI_STDIO 1
# endif
# include <windows.h>
# include <intrin.h>
# define DECLARE_WINDOWS_MAIN_FUNCTION                                 \
         TRANSPROG_API int _windowsMainFunction(int (*)(int,char **)); \
         int                                                           \
         WINAPI                                                        \
         WinMain(HINSTANCE hInstance,                                  \
                 HINSTANCE hPrevInstance,                              \
                 LPSTR lpCmdLine,                                      \
                 int nCmdShow)                                         \
         {                                                             \
         (void)hInstance;                                              \
         (void)hPrevInstance;                                          \
         (void)lpCmdLine;                                              \
         (void)nCmdShow;                                               \
         return _windowsMainFunction(&main);                           \
         }
  struct _PlatformW32Fncts
  {
  HMODULE mainModule;
  HMODULE kernel32;
  SIZE_T
  (WINAPI *GetLargePageMinimum)(VOID);
  BOOL
  (WINAPI *GetLogicalProcessorInformation)
          (SYSTEM_LOGICAL_PROCESSOR_INFORMATION *,
           DWORD *);
  BOOL
  (WINAPI *GetLogicalProcessorInformationEx)
          (LOGICAL_PROCESSOR_RELATIONSHIP,
           SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *,
           DWORD *);
  BOOL
  (WINAPI *SetThreadGroupAffinity)
          (HANDLE,
           const GROUP_AFFINITY *,
           PGROUP_AFFINITY);
  VOID
  (WINAPI *InitializeConditionVariable)
          (CONDITION_VARIABLE *);
  BOOL
  (WINAPI *SleepConditionVariableCS)
          (CONDITION_VARIABLE *,
           CRITICAL_SECTION *,
           DWORD);
  VOID
  (WINAPI *WakeAllConditionVariable)
          (CONDITION_VARIABLE *);
  };
#endif

#if !defined _MSC_VER
# if !defined _REENTRANT
#   error "_REENTRANT undefined, use -D_REENTRANT when compiling!"
# endif
# define _FILE_OFFSET_BITS 64 // large file support on 32 bit OS
# if !defined __STDC_LIMIT_MACROS
#   define __STDC_LIMIT_MACROS 1
# endif
# if !defined __STDC_CONSTANT_MACROS
#   define __STDC_CONSTANT_MACROS 1
# endif
# if !defined __STDC_FORMAT_MACROS
#   define __STDC_FORMAT_MACROS 1
# endif
# include <inttypes.h>
# include <limits.h>
  typedef float real32_t;
  typedef __attribute__((__aligned__(8))) double real64_t;
# include <sys/types.h>
# if defined _WIN32
#   include <sys/param.h>
# else // not _WIN32
#   include <sys/wait.h>
#   include <sys/stat.h>
#   include <fcntl.h>
#   include <dirent.h>
#   include <sys/param.h>
#   include <sys/mman.h>
#   if !defined MAP_ANONYMOUS && defined MAP_ANON
#     define MAP_ANONYMOUS MAP_ANON
#   endif
#   if defined __APPLE__
#     include <AvailabilityMacros.h>
#     if defined AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER
#       include <mach/mach.h>
#       include <mach/thread_policy.h>
#     endif
#   endif
#   if defined __linux__
#     include <linux/mman.h> // required for MAP_HUGETLB
#     if !defined __USE_GNU
#       define __USE_GNU // required for CPU affinity
#       include <pthread.h>
#       include <sched.h>
#       undef __USE_GNU
#     else
#       include <pthread.h>
#       include <sched.h>
#     endif
#   else
#     include <pthread.h>
#   endif
#   if defined __FreeBSD__
#     include <sys/cpuset.h>
#   endif
#   if defined __OpenBSD__
#   endif
#   if defined __NetBSD__
#   endif
#   include <unistd.h>
#   include <sys/sysctl.h>
#   include <time.h>
#   include <sys/time.h>
#   include <errno.h>
#   define RESTART_SYSCALL(result,syscall) \
           do { (result)=(syscall); } while(((result)<0)&&(errno==EINTR))
#   if !defined __APPLE__
#     define TRANSPROG_THREAD_STORAGE __thread
#   endif
#   define DECLARE_WINDOWS_MAIN_FUNCTION
# endif
#endif

#if !defined __cplusplus
#  include <stdbool.h>
#endif
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stddef.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#if !defined BYTE_ORDER
# error "BYTE_ORDER not defined, please check platform settings"
#endif

#if !defined PATH_MAX
# if defined MAX_PATH
#   define PATH_MAX MAX_PATH
# else
#   define PATH_MAX 4096 // assumed default value
# endif
#endif

// 64-bit integer printf/scanf format
#if defined _MSC_VER
# define F64 "I64"
#elif defined __APPLE__
# define F64 "ll"
#elif LONG_MAX==INT_MAX
# define F64 "ll"
#else
# define F64 "l"
#endif

// pointer integer printf/scanf format
#if defined __APPLE__
# define FPTR "l"
#elif INTPTR_MAX==INT64_MAX
# define FPTR F64
#elif INTPTR_MAX==INT32_MAX
# define FPTR ""
#else
# error "weird size for intptr_t!"
#endif

//
// Note on the usage of integer types:
//
// A signed int should be used as general purpose counter.
// It as generaly the faster integer for many arithmetic operations.
// As it is signed, many optimizations, such as auto-vectorization, will not
// consider any potential overflow and will be enabled.
//
// To count an amount of data in memory which could overflow the with of int,
// intptr_t has to be used instead. It maps to int32_t or int64_t depending on
// the size of the address-space.
//
// For an amount of data even bigger (a time in microseconds, a file size, an
// amount of physical memory greater than the 32-bit address-space...) int64_t
// will be wide enough.
//
// In any case, we avoid using unsigned integers since they are sometimes
// slower than their signed counterpart and can prevent some optimizations
// from being enabled.
//
// The only exception is when many integer values have to be stored in arrays
// or structures. We then consider using narrower integers to save space in
// order to lower cache pressure.
//

#define ONE_IF_ZERO(x) \
        ((x)+!(x))

#define OFFSET_OF_MEMBER(struct_type,member_name) \
        ((int)(intptr_t)&((struct_type *)0)->member_name)

#define STRUCT_OF_MEMBER(member_addr,struct_type,member_name)        \
        ((struct_type *)(((uint8_t *)member_addr)-                   \
                         OFFSET_OF_MEMBER(struct_type,member_name)))

#define PADDED_ELEMENT(base_addr,elem_index,elem_size,elem_type) \
        ((elem_type *)(((uint8_t *)(base_addr))+(elem_index)*(elem_size)))

#define DROP_DEAD(msg)                                    \
        do                                                \
          {                                               \
          Platform_msg("\n%s:%d:%s()\n  %s\n",            \
                       __FILE__,__LINE__,__func__,(msg)); \
          Platform_dropDead();                            \
          } while(0)

#define SYS_DROP_DEAD2(msg,err)                                    \
        do                                                         \
          {                                                        \
          char __errBuf[0x100];                                    \
          Platform_errorMsg((err),__errBuf,0x100);                 \
          Platform_msg("\n%s:%d:%s()\n  %s --- %s\n",              \
                       __FILE__,__LINE__,__func__,(msg),__errBuf); \
          Platform_dropDead();                                     \
          } while(0)

#define SYS_DROP_DEAD(msg) \
        SYS_DROP_DEAD2(msg,Platform_lastError())

#if defined NDEBUG
# define ASSERT_CURRENT_CPUCTX(cpuCtx) \
         do { (void)(cpuCtx); } while(0)
# define ON_ERROR(condition,msg,action) \
         do { (void)(condition); } while(0)
#else
# define ASSERT_CURRENT_CPUCTX(cpuCtx)          \
         do                                     \
           {                                    \
           if((cpuCtx)!=Platform_cpuCtx())      \
             {                                  \
             DROP_DEAD("cpu context mismatch"); \
             }                                  \
           } while(0)
# define ON_ERROR(condition,msg,action)                      \
         do                                                  \
           {                                                 \
           if((condition))                                   \
             {                                               \
             Platform_msg("\n%s:%d:%s()\n  %s\n",            \
                          __FILE__,__LINE__,__func__,(msg)); \
             action;                                         \
             }                                               \
           } while(0)
#endif

#if defined _MSC_VER
# define CALL_CPUID(level,arg,a,b,c,d)      \
         do                                 \
           {                                \
           int __info[4];                   \
           __cpuidex(__info,(level),(arg)); \
           (a)=__info[0];                   \
           (b)=__info[1];                   \
           (c)=__info[2];                   \
           (d)=__info[3];                   \
           } while(0)
#else
# if defined __i386__
#   define CALL_CPUID(level,arg,a,b,c,d)                              \
           __asm__ __volatile__("\n"                                  \
           "pushl %%ebx\n\t"   /* save ebx (in case of PIC) */        \
           "cpuid\n\t"                                                \
           "movl %%ebx,%1\n\t" /* store ebx content to esi */         \
           "popl %%ebx\n\t"    /* restore ebx */                      \
           : "=a"(a),"=S"(b),"=c"(c),"=d"(d) : "a"(level),"c"(arg) )
# elif defined __x86_64__
#   define CALL_CPUID(level,arg,a,b,c,d)                              \
           __asm__ __volatile__("\n"                                  \
           "pushq %%rbx\n\t"   /* save rbx (in case of PIC) */        \
           "cpuid\n\t"                                                \
           "movl %%ebx,%1\n\t" /* store ebx content to esi */         \
           "popq %%rbx\n\t"    /* restore rbx */                      \
           : "=a"(a),"=S"(b),"=c"(c),"=d"(d) : "a"(level),"c"(arg) )
# endif
#endif

#define ASSUMED_MAX_CACHE_LEVEL 3   // adjust as needed
#define ASSUMED_CACHE_ALIGNMENT 64  // adjust as needed
#define ASSUMED_MAX_NUMA_COUNT  4   // adjust as needed
#define ASSUMED_MAX_MEM_COUNT   512 // adjust as needed
#define ASSUMED_MAX_CPU_COUNT   256 // adjust as needed

#if defined _MSC_VER
# define CACHE_ALIGNED __declspec(align(ASSUMED_CACHE_ALIGNMENT))
#else
# define CACHE_ALIGNED __attribute__((__aligned__(ASSUMED_CACHE_ALIGNMENT)))
#endif

#if (ASSUMED_MAX_CPU_COUNT-1)<=UINT8_MAX
  typedef uint8_t NarrowCpuCtx;
#elif (ASSUMED_MAX_CPU_COUNT-1)<=UINT16_MAX
  typedef uint16_t NarrowCpuCtx;
#endif

struct _CpumemInfoNode
{
CACHE_ALIGNED
int parent;
int firstChild;
int childCount;
int firstCpu;
int cpuCount;
int numa;
int cpuId;
int cacheSize[ASSUMED_MAX_CACHE_LEVEL];
int cacheLineSize[ASSUMED_MAX_CACHE_LEVEL];
};

struct _CpuMemInfo
{
const char *cpuMask;
bool flattenCpu;
bool attachCpu;
int64_t memSize; // can exceed address-space on 32-bit systems
intptr_t pageSize;
intptr_t hugePageSize;
int maxCacheLineSize;
int maxCacheLevel;
int maxCpuId;
int maxNumaId;
int memCount;
int cpuCount;
int numaCount;
int numaId[1+ASSUMED_MAX_NUMA_COUNT];
struct _CpumemInfoNode nodes[ASSUMED_MAX_MEM_COUNT];
int maxCpuDistance;
struct
  {
  CACHE_ALIGNED
  uint8_t distances[ASSUMED_MAX_CPU_COUNT]; // minimize space
  } cpuDistances[ASSUMED_MAX_CPU_COUNT];
struct
  {
  CACHE_ALIGNED
  uint8_t proximities[ASSUMED_MAX_CPU_COUNT]; // minimize space
  } cpuProximities[ASSUMED_MAX_CPU_COUNT];
struct
  {
  CACHE_ALIGNED
  NarrowCpuCtx trip[ASSUMED_MAX_CPU_COUNT]; // minimize space
  } cpuRoundTrips[ASSUMED_MAX_CPU_COUNT];
};

TRANSPROG_API
void
Platform_init(int argc,
              char **argv);

TRANSPROG_INLINE
int // last error code
Platform_lastError(void)
{
#if defined _WIN32
  return GetLastError();
#else
  return errno;
#endif
}

TRANSPROG_API
void
Platform_errorMsg(int errorCode,
                  char *msg,
                  int msgSize);

TRANSPROG_INLINE
void
Platform_lastErrorMsg(char *msg,
                      int msgSize)
{
Platform_errorMsg(Platform_lastError(),msg,msgSize);
}

TRANSPROG_API
int
Platform_msg(const char *format,
             ...)
#if !defined _MSC_VER
# if defined _WIN32
    __attribute__((__format__(gnu_printf,1,2)))
# else
    __attribute__((__format__(printf,1,2)))
# endif
#endif
;

TRANSPROG_API
void
Platform_dropDead(void);

TRANSPROG_INLINE
int // number of command line arguments
Platform_argc(void)
{
extern TRANSPROG_API_VAR int _Platform_argc;
return _Platform_argc;
}

TRANSPROG_INLINE
const char * // index'th command line argument
Platform_argv(int index)
{
extern TRANSPROG_API_VAR char **_Platform_argv;
return _Platform_argv[index];
}

TRANSPROG_INLINE
const char * // current executable path
Platform_executable(void)
{
extern TRANSPROG_API_VAR char *_Platform_executable;
return _Platform_executable;
}

TRANSPROG_API
bool // success
Platform_environ(const char *varName,
                 char *out_value,
                 int size);

TRANSPROG_API
int // keyword index
Platform_environAsKeyword(const char *varName,
                          const char * const *keywords,
                          int keywordCount,
                          int defaultIndex);

TRANSPROG_API
bool // boolean value extracted from environment variable
Platform_environAsBool(const char *varName,
                       bool defaultValue);

TRANSPROG_API
int // integer value extracted from environment variable
Platform_environAsInt(const char *varName,
                      int defaultValue);

TRANSPROG_API
int64_t // microseconds since 1970/01/01 UTC
Platform_time(void);

TRANSPROG_API
void
Platform_sleep(int64_t microseconds);

TRANSPROG_INLINE
void
Platform_pause(void)
{
#if defined _MSC_VER
# if defined _M_IX86 || defined __INTEL_COMPILER
    __asm { pause }
# else
    __nop();
# endif
#elif defined __POWERPC__
   __asm__ __volatile__ ("ori r0,r0,0"); // fake nop
#elif defined __i386__ || defined __x86_64__
  __asm__ __volatile__ ("pause");
#elif 0 // may not be available everywhere
  __builtin_ia32_pause();
#else
  // just do nothing...
#endif
}

TRANSPROG_INLINE
void
Platform_prefetch(const void *addr)
{
#if defined _MSC_VER
  _mm_prefetch((char *)addr,_MM_HINT_T0);
#else
  __builtin_prefetch(addr);
#endif
}

TRANSPROG_INLINE
int // current cpu context
Platform_cpuCtx(void)
{
void *cpuData;
#if defined TRANSPROG_THREAD_STORAGE
  extern TRANSPROG_API_VAR TRANSPROG_THREAD_STORAGE void *_Platform_cpuData;
  cpuData=_Platform_cpuData;
#elif defined _WIN32
  extern TRANSPROG_API_VAR DWORD _Platform_cpuData;
  cpuData=TlsGetValue(_Platform_cpuData);
#else
  extern TRANSPROG_API_VAR pthread_key_t _Platform_cpuData;
  cpuData=pthread_getspecific(_Platform_cpuData);
#endif
return (int)(intptr_t)cpuData;
}

TRANSPROG_INLINE
int // max system cpu id or -1 if unknown
Platform_maxCpuId(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCpuId;
}

TRANSPROG_INLINE
int // number of cpu contexts
Platform_cpuCtxCount(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuCount;
}

TRANSPROG_INLINE
int // system cpu id of memory context or -1 if not a cpu
Platform_cpuId(int cpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[cpuCtx].cpuId;
}

TRANSPROG_INLINE
int // max cache distance from one cpu to another
Platform_maxCpuDistance(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCpuDistance;
}

TRANSPROG_INLINE
int // cache distance from one cpu to another
Platform_cpuDistance(int cpuCtx,
                     int otherCpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuDistances[cpuCtx].distances[otherCpuCtx];
}

TRANSPROG_INLINE
int // cache proximity from one cpu to another
Platform_cpuProximity(int cpuCtx,
                      int otherCpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuProximities[cpuCtx].proximities[otherCpuCtx];
}

TRANSPROG_INLINE
const NarrowCpuCtx * // cache-aware round-trip starting from a cpu
Platform_cpuRoundTrip(int cpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuRoundTrips[cpuCtx].trip;
}

TRANSPROG_INLINE
int // max system numa id or -1 if unknown
Platform_maxNumaId(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxNumaId;
}

TRANSPROG_INLINE
int // number of numa contexts
Platform_numaCtxCount(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.numaCount;
}

TRANSPROG_INLINE
int // system id of numa context or -1 if unknown
Platform_numaId(int numaCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.numaId[numaCtx];
}

TRANSPROG_INLINE
int // max cache level or 0 if unknown
Platform_maxCacheLevel(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCacheLevel;
}

TRANSPROG_INLINE
int // max cache line size
Platform_maxCacheLineSize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCacheLineSize;
}

TRANSPROG_INLINE
int // root memory context
Platform_rootMemCtx(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.memCount-1;
}

TRANSPROG_INLINE
int // number of memory contexts
Platform_memCtxCount(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.memCount;
}

TRANSPROG_INLINE
int // numa context of memory context
Platform_memNumaCtx(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].numa;
}

TRANSPROG_INLINE
int // parent of memory context or -1 if none
Platform_memParent(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].parent;
}

TRANSPROG_INLINE
int // number of children in memory context
Platform_memChildCount(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].childCount;
}

TRANSPROG_INLINE
int // first child of memory context
Platform_memFirstChild(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].firstChild;
}

TRANSPROG_INLINE
int // number of cpu contexts in memory context
Platform_memCpuCount(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].cpuCount;
}

TRANSPROG_INLINE
int // first cpu context of memory context
Platform_memFirstCpu(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].firstCpu;
}

TRANSPROG_INLINE
int // cache size in bytes of memory context or 0 if unavailable
Platform_memCacheSize(int memCtx,
                      int level)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return (level>0)&&(level<=_Platform_cpuMemInfo.maxCacheLevel)
       ? _Platform_cpuMemInfo.nodes[memCtx].cacheSize[level-1]
       : 0;
}

TRANSPROG_INLINE
int // cache line size in bytes of memory context or 0 if unavailable
Platform_memCacheLineSize(int memCtx,
                          int level)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return (level>0)&&(level<=_Platform_cpuMemInfo.maxCacheLevel)
       ? _Platform_cpuMemInfo.nodes[memCtx].cacheLineSize[level-1]
       : 0;
}

TRANSPROG_INLINE
int64_t // system memory size in bytes or 0 if unknown
Platform_memorySize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.memSize;
}

TRANSPROG_INLINE
intptr_t // standard system memory page size in bytes
Platform_standardPageSize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.pageSize;
}

TRANSPROG_INLINE
intptr_t // huge system memory page size in bytes or 0 if unavailable
Platform_hugePageSize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.hugePageSize;
}

TRANSPROG_INLINE
intptr_t // system memory page size in bytes
Platform_pageSize(bool preferHuge)
{
intptr_t size=preferHuge ? Platform_hugePageSize() : 0;
return size ? size : Platform_standardPageSize();
}

TRANSPROG_INLINE
intptr_t // size rounded to next alignment
Platform_alignSize(intptr_t size,
                   intptr_t alignment)
{
size+=alignment-1;
return size-(size%alignment);
}

TRANSPROG_INLINE
void * // address rounded to next alignment
Platform_alignAddress(void *address,
                      intptr_t alignment)
{
intptr_t addr=(intptr_t)address;
addr+=alignment-1;
return (void *)(addr-(addr%alignment));
}

TRANSPROG_INLINE
intptr_t // begin index of part splitIndex of [begin;end[ cut in splitCount
Platform_splitBegin(intptr_t begin,
                    intptr_t end,
                    int splitCount,
                    int splitIndex)
{
int64_t range=end-begin;
intptr_t splitBegin=begin+(intptr_t)((range*splitIndex)/splitCount);
return splitBegin<end ? splitBegin : end;
}

TRANSPROG_INLINE
intptr_t // end index of part splitIndex of [begin;end[ cut in splitCount
Platform_splitEnd(intptr_t begin,
                  intptr_t end,
                  int splitCount,
                  int splitIndex)
{
return Platform_splitBegin(begin,end,splitCount,splitIndex+1);
}

TRANSPROG_API
void
Platform_parallel(void (*fnct)(int,void *),
                  void *data);

#endif // TRANSPROG_PLATFORM_H

//----------------------------------------------------------------------------
