//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_SURFACE_H
#define TRANSPROG_SURFACE_H 1

#include "TransProg/platform.h"

TRANSPROG_API
void
Surface_boxRequirements(int *out_indexCount,
                        int *out_vertexCount);

TRANSPROG_API
void
Surface_makeBoxGeneric(void *indices,
                       int sizeOfIndex,
                       void *vertices,
                       int sizeOfVertex,
                       void *normals,
                       int sizeOfNormal,
                       void *texCoords,
                       int sizeOfTexCoord,
                       bool counterClockWise,
                       real64_t sizeX,
                       real64_t sizeY,
                       real64_t sizeZ);

#define Surface_makeBox(indices,                          \
                        vertices,                         \
                        normals,                          \
                        texCoords,                        \
                        counterClockWise,                 \
                        sizeX,                            \
                        sizeY,                            \
                        sizeZ)                            \
        Surface_makeBoxGeneric((indices),                 \
                               (int)sizeof(*(indices)),   \
                               (vertices),                \
                               (int)sizeof(*(vertices)),  \
                               (normals),                 \
                               (int)sizeof(*(normals)),   \
                               (texCoords),               \
                               (int)sizeof(*(texCoords)), \
                               (counterClockWise),        \
                               (sizeX),                   \
                               (sizeY),                   \
                               (sizeZ))

TRANSPROG_API
void
Surface_allocBoxGeneric(int *out_indexCount,
                        void **indices,
                        int sizeOfIndex,
                        int *out_vertexCount,
                        void **vertices,
                        int sizeOfVertex,
                        void **normals,
                        int sizeOfNormal,
                        void **texCoords,
                        int sizeOfTexCoord,
                        bool counterClockWise,
                        real64_t sizeX,
                        real64_t sizeY,
                        real64_t sizeZ);

#define Surface_allocBox(out_indexCount,                    \
                         indices,                           \
                         out_vertexCount,                   \
                         vertices,                          \
                         normals,                           \
                         texCoords,                         \
                         counterClockWise,                  \
                         sizeX,                             \
                         sizeY,                             \
                         sizeZ)                             \
        Surface_allocBoxGeneric((out_indexCount),           \
                                (void **)(indices),         \
                                (int)sizeof(**(indices)),   \
                                (out_vertexCount),          \
                                (void **)(vertices),        \
                                (int)sizeof(**(vertices)),  \
                                (void **)(normals),         \
                                (int)sizeof(**(normals)),   \
                                (void **)(texCoords),       \
                                (int)sizeof(**(texCoords)), \
                                (counterClockWise),         \
                                (sizeX),                    \
                                (sizeY),                    \
                                (sizeZ))

TRANSPROG_API
void
Surface_sphereRequirements(int *out_indexCount,
                           int *out_vertexCount,
                           int precision);

TRANSPROG_API
void
Surface_makeSphereGeneric(void *indices,
                          int sizeOfIndex,
                          void *vertices,
                          int sizeOfVertex,
                          void *normals,
                          int sizeOfNormal,
                          void *texCoords,
                          int sizeOfTexCoord,
                          bool counterClockWise,
                          real64_t radius,
                          int precision);

#define Surface_makeSphere(indices,                          \
                           vertices,                         \
                           normals,                          \
                           texCoords,                        \
                           counterClockWise,                 \
                           radius,                           \
                           precision)                        \
        Surface_makeSphereGeneric((indices),                 \
                                  (int)sizeof(*(indices)),   \
                                  (vertices),                \
                                  (int)sizeof(*(vertices)),  \
                                  (normals),                 \
                                  (int)sizeof(*(normals)),   \
                                  (texCoords),               \
                                  (int)sizeof(*(texCoords)), \
                                  (counterClockWise),        \
                                  (radius),                  \
                                  (precision))

TRANSPROG_API
void
Surface_allocSphereGeneric(int *out_indexCount,
                           void **indices,
                           int sizeOfIndex,
                           int *out_vertexCount,
                           void **vertices,
                           int sizeOfVertex,
                           void **normals,
                           int sizeOfNormal,
                           void **texCoords,
                           int sizeOfTexCoord,
                           bool counterClockWise,
                           real64_t radius,
                           int precision);

#define Surface_allocSphere(out_indexCount,                    \
                            indices,                           \
                            out_vertexCount,                   \
                            vertices,                          \
                            normals,                           \
                            texCoords,                         \
                            counterClockWise,                  \
                            radius,                            \
                            precision)                         \
        Surface_allocSphereGeneric((out_indexCount),           \
                                   (void **)(indices),         \
                                   (int)sizeof(**(indices)),   \
                                   (out_vertexCount),          \
                                   (void **)(vertices),        \
                                   (int)sizeof(**(vertices)),  \
                                   (void **)(normals),         \
                                   (int)sizeof(**(normals)),   \
                                   (void **)(texCoords),       \
                                   (int)sizeof(**(texCoords)), \
                                   (counterClockWise),         \
                                   (radius),                   \
                                   (precision))

TRANSPROG_API
void
Surface_ellipsoidRequirements(int *out_indexCount,
                              int *out_vertexCount,
                              int precision);

TRANSPROG_API
void
Surface_makeEllipsoidGeneric(void *indices,
                             int sizeOfIndex,
                             void *vertices,
                             int sizeOfVertex,
                             void *normals,
                             int sizeOfNormal,
                             void *texCoords,
                             int sizeOfTexCoord,
                             bool counterClockWise,
                             real64_t radiusX,
                             real64_t radiusY,
                             real64_t radiusZ,
                             int precision);

#define Surface_makeEllipsoid(indices,                       \
                              vertices,                      \
                              normals,                       \
                              texCoords,                     \
                              counterClockWise,              \
                              radiusX,                       \
                              radiusY,                       \
                              radiusZ,                       \
                              precision)                     \
        Surface_makeEllipsoidGeneric((indices),              \
                                  (int)sizeof(*(indices)),   \
                                  (vertices),                \
                                  (int)sizeof(*(vertices)),  \
                                  (normals),                 \
                                  (int)sizeof(*(normals)),   \
                                  (texCoords),               \
                                  (int)sizeof(*(texCoords)), \
                                  (counterClockWise),        \
                                  (radiusX),                 \
                                  (radiusY),                 \
                                  (radiusZ),                 \
                                  (precision))

TRANSPROG_API
void
Surface_allocEllipsoidGeneric(int *out_indexCount,
                              void **indices,
                              int sizeOfIndex,
                              int *out_vertexCount,
                              void **vertices,
                              int sizeOfVertex,
                              void **normals,
                              int sizeOfNormal,
                              void **texCoords,
                              int sizeOfTexCoord,
                              bool counterClockWise,
                              real64_t radiusX,
                              real64_t radiusY,
                              real64_t radiusZ,
                              int precision);

#define Surface_allocEllipsoid(out_indexCount,                 \
                               indices,                        \
                               out_vertexCount,                \
                               vertices,                       \
                               normals,                        \
                               texCoords,                      \
                               counterClockWise,               \
                               radiusX,                        \
                               radiusY,                        \
                               radiusZ,                        \
                               precision)                      \
        Surface_allocEllipsoidGeneric((out_indexCount),        \
                                   (void **)(indices),         \
                                   (int)sizeof(**(indices)),   \
                                   (out_vertexCount),          \
                                   (void **)(vertices),        \
                                   (int)sizeof(**(vertices)),  \
                                   (void **)(normals),         \
                                   (int)sizeof(**(normals)),   \
                                   (void **)(texCoords),       \
                                   (int)sizeof(**(texCoords)), \
                                   (counterClockWise),         \
                                   (radiusX),                  \
                                   (radiusY),                  \
                                   (radiusZ),                  \
                                   (precision))

TRANSPROG_API
void
Surface_coneRequirements(int *out_indexCount,
                         int *out_vertexCount,
                         bool side,
                         bool bottom,
                         int precision);

TRANSPROG_API
void
Surface_makeConeGeneric(void *indices,
                        int sizeOfIndex,
                        void *vertices,
                        int sizeOfVertex,
                        void *normals,
                        int sizeOfNormal,
                        void *texCoords,
                        int sizeOfTexCoord,
                        bool counterClockWise,
                        real64_t radius,
                        real64_t height,
                        bool side,
                        bool bottom,
                        int precision);

#define Surface_makeCone(indices,                          \
                         vertices,                         \
                         normals,                          \
                         texCoords,                        \
                         counterClockWise,                 \
                         radius,                           \
                         height,                           \
                         side,                             \
                         bottom,                           \
                         precision)                        \
        Surface_makeConeGeneric((indices),                 \
                                (int)sizeof(*(indices)),   \
                                (vertices),                \
                                (int)sizeof(*(vertices)),  \
                                (normals),                 \
                                (int)sizeof(*(normals)),   \
                                (texCoords),               \
                                (int)sizeof(*(texCoords)), \
                                (counterClockWise),        \
                                (radius),                  \
                                (height),                  \
                                (side),                    \
                                (bottom),                  \
                                (precision))

TRANSPROG_API
void
Surface_allocConeGeneric(int *out_indexCount,
                         void **indices,
                         int sizeOfIndex,
                         int *out_vertexCount,
                         void **vertices,
                         int sizeOfVertex,
                         void **normals,
                         int sizeOfNormal,
                         void **texCoords,
                         int sizeOfTexCoord,
                         bool counterClockWise,
                         real64_t radius,
                         real64_t height,
                         bool side,
                         bool bottom,
                         int precision);

#define Surface_allocCone(out_indexCount,                    \
                          indices,                           \
                          out_vertexCount,                   \
                          vertices,                          \
                          normals,                           \
                          texCoords,                         \
                          counterClockWise,                  \
                          radius,                            \
                          height,                            \
                          side,                              \
                          bottom,                            \
                          precision)                         \
        Surface_allocConeGeneric((out_indexCount),           \
                                 (void **)(indices),         \
                                 (int)sizeof(**(indices)),   \
                                 (out_vertexCount),          \
                                 (void **)(vertices),        \
                                 (int)sizeof(**(vertices)),  \
                                 (void **)(normals),         \
                                 (int)sizeof(**(normals)),   \
                                 (void **)(texCoords),       \
                                 (int)sizeof(**(texCoords)), \
                                 (counterClockWise),         \
                                 (radius),                   \
                                 (height),                   \
                                 (side),                     \
                                 (bottom),                   \
                                 (precision))

TRANSPROG_API
void
Surface_cylinderRequirements(int *out_indexCount,
                             int *out_vertexCount,
                             bool side,
                             bool top,
                             bool bottom,
                             int precision);

TRANSPROG_API
void
Surface_makeCylinderGeneric(void *indices,
                            int sizeOfIndex,
                            void *vertices,
                            int sizeOfVertex,
                            void *normals,
                            int sizeOfNormal,
                            void *texCoords,
                            int sizeOfTexCoord,
                            bool counterClockWise,
                            real64_t radius,
                            real64_t height,
                            bool side,
                            bool top,
                            bool bottom,
                            int precision);

#define Surface_makeCylinder(indices,                          \
                             vertices,                         \
                             normals,                          \
                             texCoords,                        \
                             counterClockWise,                 \
                             radius,                           \
                             height,                           \
                             side,                             \
                             top,                              \
                             bottom,                           \
                             precision)                        \
        Surface_makeCylinderGeneric((indices),                 \
                                    (int)sizeof(*(indices)),   \
                                    (vertices),                \
                                    (int)sizeof(*(vertices)),  \
                                    (normals),                 \
                                    (int)sizeof(*(normals)),   \
                                    (texCoords),               \
                                    (int)sizeof(*(texCoords)), \
                                    (counterClockWise),        \
                                    (radius),                  \
                                    (height),                  \
                                    (side),                    \
                                    (top),                     \
                                    (bottom),                  \
                                    (precision))

TRANSPROG_API
void
Surface_allocCylinderGeneric(int *out_indexCount,
                             void **indices,
                             int sizeOfIndex,
                             int *out_vertexCount,
                             void **vertices,
                             int sizeOfVertex,
                             void **normals,
                             int sizeOfNormal,
                             void **texCoords,
                             int sizeOfTexCoord,
                             bool counterClockWise,
                             real64_t radius,
                             real64_t height,
                             bool side,
                             bool top,
                             bool bottom,
                             int precision);

#define Surface_allocCylinder(out_indexCount,                    \
                              indices,                           \
                              out_vertexCount,                   \
                              vertices,                          \
                              normals,                           \
                              texCoords,                         \
                              counterClockWise,                  \
                              radius,                            \
                              height,                            \
                              side,                              \
                              top,                               \
                              bottom,                            \
                              precision)                         \
        Surface_allocCylinderGeneric((out_indexCount),           \
                                     (void **)(indices),         \
                                     (int)sizeof(**(indices)),   \
                                     (out_vertexCount),          \
                                     (void **)(vertices),        \
                                     (int)sizeof(**(vertices)),  \
                                     (void **)(normals),         \
                                     (int)sizeof(**(normals)),   \
                                     (void **)(texCoords),       \
                                     (int)sizeof(**(texCoords)), \
                                     (counterClockWise),         \
                                     (radius),                   \
                                     (height),                   \
                                     (side),                     \
                                     (top),                      \
                                     (bottom),                   \
                                     (precision))

TRANSPROG_API
void
Surface_torusRequirements(int *out_indexCount,
                          int *out_vertexCount,
                          int precision1,
                          int precision2);

TRANSPROG_API
void
Surface_makeTorusGeneric(void *indices,
                         int sizeOfIndex,
                         void *vertices,
                         int sizeOfVertex,
                         void *normals,
                         int sizeOfNormal,
                         void *texCoords,
                         int sizeOfTexCoord,
                         bool counterClockWise,
                         real64_t radius1,
                         real64_t radius2,
                         int precision1,
                         int precision2);

#define Surface_makeTorus(indices,                          \
                          vertices,                         \
                          normals,                          \
                          texCoords,                        \
                          counterClockWise,                 \
                          radius1,                          \
                          radius2,                          \
                          precision1,                       \
                          precision2)                       \
        Surface_makeTorusGeneric((indices),                 \
                                 (int)sizeof(*(indices)),   \
                                 (vertices),                \
                                 (int)sizeof(*(vertices)),  \
                                 (normals),                 \
                                 (int)sizeof(*(normals)),   \
                                 (texCoords),               \
                                 (int)sizeof(*(texCoords)), \
                                 (counterClockWise),        \
                                 (radius1),                 \
                                 (radius2),                 \
                                 (precision1),              \
                                 (precision2))

TRANSPROG_API
void
Surface_allocTorusGeneric(int *out_indexCount,
                          void **indices,
                          int sizeOfIndex,
                          int *out_vertexCount,
                          void **vertices,
                          int sizeOfVertex,
                          void **normals,
                          int sizeOfNormal,
                          void **texCoords,
                          int sizeOfTexCoord,
                          bool counterClockWise,
                          real64_t radius1,
                          real64_t radius2,
                          int precision1,
                          int precision2);

#define Surface_allocTorus(out_indexCount,                    \
                           indices,                           \
                           out_vertexCount,                   \
                           vertices,                          \
                           normals,                           \
                           texCoords,                         \
                           counterClockWise,                  \
                           radius1,                           \
                           radius2,                           \
                           precision1,                        \
                           precision2)                        \
        Surface_allocTorusGeneric((out_indexCount),           \
                                  (void **)(indices),         \
                                  (int)sizeof(**(indices)),   \
                                  (out_vertexCount),          \
                                  (void **)(vertices),        \
                                  (int)sizeof(**(vertices)),  \
                                  (void **)(normals),         \
                                  (int)sizeof(**(normals)),   \
                                  (void **)(texCoords),       \
                                  (int)sizeof(**(texCoords)), \
                                  (counterClockWise),         \
                                  (radius1),                  \
                                  (radius2),                  \
                                  (precision1),               \
                                  (precision2))

TRANSPROG_API
void
Surface_gridRequirements(int *out_indexCount,
                         int *out_vertexCount,
                         int precisionX,
                         int precisionY);

TRANSPROG_API
void
Surface_makeGridGeneric(void *indices,
                        int sizeOfIndex,
                        void *vertices,
                        int sizeOfVertex,
                        void *normals,
                        int sizeOfNormal,
                        void *texCoords,
                        int sizeOfTexCoord,
                        bool counterClockWise,
                        real64_t sizeX,
                        real64_t sizeY,
                        int precisionX,
                        int precisionY);

#define Surface_makeGrid(indices,                          \
                         vertices,                         \
                         normals,                          \
                         texCoords,                        \
                         counterClockWise,                 \
                         sizeX,                            \
                         sizeY,                            \
                         precisionX,                       \
                         precisionY)                       \
        Surface_makeGridGeneric((indices),                 \
                                (int)sizeof(*(indices)),   \
                                (vertices),                \
                                (int)sizeof(*(vertices)),  \
                                (normals),                 \
                                (int)sizeof(*(normals)),   \
                                (texCoords),               \
                                (int)sizeof(*(texCoords)), \
                                (counterClockWise),        \
                                (sizeX),                   \
                                (sizeY),                   \
                                (precisionX),              \
                                (precisionY))

TRANSPROG_API
void
Surface_allocGridGeneric(int *out_indexCount,
                         void **indices,
                         int sizeOfIndex,
                         int *out_vertexCount,
                         void **vertices,
                         int sizeOfVertex,
                         void **normals,
                         int sizeOfNormal,
                         void **texCoords,
                         int sizeOfTexCoord,
                         bool counterClockWise,
                         real64_t sizeX,
                         real64_t sizeY,
                         int precisionX,
                         int precisionY);

#define Surface_allocGrid(out_indexCount,                    \
                          indices,                           \
                          out_vertexCount,                   \
                          vertices,                          \
                          normals,                           \
                          texCoords,                         \
                          counterClockWise,                  \
                          sizeX,                             \
                          sizeY,                             \
                          precisionX,                        \
                          precisionY)                        \
        Surface_allocGridGeneric((out_indexCount),           \
                                 (void **)(indices),         \
                                 (int)sizeof(**(indices)),   \
                                 (out_vertexCount),          \
                                 (void **)(vertices),        \
                                 (int)sizeof(**(vertices)),  \
                                 (void **)(normals),         \
                                 (int)sizeof(**(normals)),   \
                                 (void **)(texCoords),       \
                                 (int)sizeof(**(texCoords)), \
                                 (counterClockWise),         \
                                 (sizeX),                    \
                                 (sizeY),                    \
                                 (precisionX),               \
                                 (precisionY))

TRANSPROG_API
void
Surface_loadMeshGeneric(const char *fileName,
                        int *out_indexCount,
                        void **out_indices,
                        int sizeOfIndex,
                        int *out_vertexCount,
                        void **out_vertices,
                        int sizeOfVertex,
                        void **out_normals,
                        int sizeOfNormal,
                        void **out_texCoords,
                        int sizeOfTexCoord);

#define Surface_loadMesh(fileName,                          \
                         out_indexCount,                    \
                         indices,                           \
                         out_vertexCount,                   \
                         vertices,                          \
                         normals,                           \
                         texCoords)                         \
        Surface_loadMeshGeneric((fileName),                 \
                                (out_indexCount),           \
                                (void **)(indices),         \
                                (int)sizeof(**(indices)),   \
                                (out_vertexCount),          \
                                (void **)(vertices),        \
                                (int)sizeof(**(vertices)),  \
                                (void **)(normals),         \
                                (int)sizeof(**(normals)),   \
                                (void **)(texCoords),       \
                                (int)sizeof(**(texCoords)))

TRANSPROG_API
void
Surface_computeDefaultNormalsGeneric(int indexCount,
                                     const void *indices,
                                     int sizeOfIndex,
                                     int vertexCount,
                                     const void *vertices,
                                     int sizeOfVertex,
                                     void *normals,
                                     int sizeOfNormal);

#define Surface_computeDefaultNormals(indexCount,                      \
                                      indices,                         \
                                      vertexCount,                     \
                                      vertices,                        \
                                      normals)                         \
        Surface_computeDefaultNormalsGeneric((indexCount),             \
                                             (indices),                \
                                             (int)sizeof(*(indices)),  \
                                             (vertexCount),            \
                                             (vertices),               \
                                             (int)sizeof(*(vertices)), \
                                             (normals),                \
                                             (int)sizeof(*(normals)))

#endif // TRANSPROG_SURFACE_H

//----------------------------------------------------------------------------
