#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import os
import sys

class Structure:
  def __init__(self,**kw):
    for (k,v) in kw.items():
      setattr(self,k,v)     

coreArbLines=[]
extensions=[]
curExt=None

def findWord(l,e,startswith=True,endswith=True):
  for i in range(len(l)):
    found=True
    if startswith and not l[i].startswith(e):
      found=False
    if endswith and not l[i].endswith(e):
      found=False
    if found:
      return i
  return -1

def findApiEntry(typeName,ext=None):
  if ext:
    ext=[ext]
  else:
    ext=extensions
  for e in ext:
    for a in e.apiEntries:
      if a.typeName==typeName:
        return a
  return None

for l in open('glcorearb.h','r'):
  coreArbLines.append(l[:])
  l=l.split()
  if len(l)==2 and l[0]=='#ifndef' and l[1].startswith('GL_'):
    extension=l[1][3:]
    if curExt:
      raise Exception('extension %s detected while in %s'
                      %(extension,curExt.name))
    curExt=Structure(name=extension,apiEntries=[])
    extensions.append(curExt)
    continue
  if len(l)==4 and l[0]=='#endif' and l[1]=='/*' and \
     l[2].startswith('GL_') and l[3]=='*/':
    extension=l[2][3:]
    if not curExt:
      raise Exception('closing extension %s out of any extension'%extension)
    if curExt.name!=extension:
      raise Exception('closing extension %s out in extension %s'
                      %(extension,curExt.name))
    curExt=None
    continue
  i=findWord(l,'typedef')
  if i==0:
    i=findWord(l,'APIENTRYP',startswith=False)
    if i!=-1 and len(l)>i+1:
      typeName=l[i+1]
      if typeName.startswith('PFNGL') and typeName.endswith('PROC)'):
        typeName=typeName[5:-5]
        if not curExt:
          raise Exception('type %s detected out of any extension'%typeName)
        if findApiEntry(typeName):
          raise Exception('multiple occurences of type %s'%typeName)
        curExt.apiEntries.append(Structure(typeName=typeName,name=None))
        continue
  i=findWord(l,'GLAPI')
  if i==0:
    i=findWord(l,'APIENTRY',startswith=False)
    if i!=-1 and len(l)>i+1:
      name=l[i+1]
      if name.startswith('gl'):
        name=name[2:]
        typeName=name.upper()
        if not curExt:
          raise Exception('entry %s detected out of any extension'%name)
        e=findApiEntry(typeName)
        if not e:
          raise Exception('no type %s for entry %s'%(typeName,name))
        if not e in curExt.apiEntries:
          raise Exception('type %s should be in extension %s'
                          %(typeName,curExt.name))
        if e.name:
          raise Exception('multiple names for type %s'%typeName)
        e.name=name
        continue

#-----------------------------------------------------------------------------

header=open('coreGl.h','w')
source=open('coreGl.c','w')

preamble='//'+'-'*76+r'''
// Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

// This Core OpenGL wrapper is generated from:
//    http://www.opengl.org/registry/api/GL/glcorearb.h
// Do not edit or your changes will be lost!
'''

header.write(preamble)
source.write(preamble)

header.write(r'''
#ifndef TRANSGL_COREGL_H
#define TRANSGL_COREGL_H 1

#include "TransGl/api.h"
//
// The preceding inclusion can be commented out in order to make this code
// independant from the TransProg library.
// In this case, the following settings are required.
// They may need some adjustments on some platforms.
//
#if !defined TRANSGL_API_H // TransProg library is not used

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <ctype.h>
# if !defined __cplusplus
#   include <stdbool.h>
# endif

# define Memory_alloc(size,type) \
         ((type *)malloc((size)*sizeof(type)))
# define Memory_initAlloc(size,type) \
         ((type *)calloc((size),sizeof(type)))
# define Memory_reAlloc(oldPtr,newSize,type) \
         ((type *)realloc((oldPtr),(newSize)*sizeof(type)))
# define Memory_free(ptr) \
         (free((ptr)))

# define TRANSGL_API_VAR // fake definition
# define TRANSGL_API // fake definition

# if !defined TRANSPROG_EXTERN_C
#   if defined __cplusplus
#     define TRANSPROG_EXTERN_C extern "C"
#   else
#     define TRANSPROG_EXTERN_C
#   endif
# endif

# if !defined TRANSPROG_INLINE
#   if defined _MSC_VER
#     define TRANSPROG_INLINE __inline static
#   else
#     define TRANSPROG_INLINE inline static
#   endif
# endif

# if defined _MSC_VER && !defined __func__
#   define __func__ __FUNCTION__
# endif

# if !defined ON_ERROR
#   if defined NDEBUG
#     define ON_ERROR(condition,msg,action) \
             do { (void)(condition); } while(0)
#   else
#     define ON_ERROR(condition,msg,action)                        \
             do                                                    \
               {                                                   \
               if((condition))                                     \
                 {                                                 \
                 fprintf(stderr,"\n%%s:%%d:%%s()\n  %%s\n",            \
                                __FILE__,__LINE__,__func__,(msg)); \
                 action;                                           \
                 }                                                 \
               } while(0)
#   endif
# endif

#endif // not TRANSGL_API_H

//%s

'''%('-'*76))
for l in coreArbLines:
  header.write(l)
header.write(r'''
//%s

#if defined NDEBUG
# define ON_GL_ERROR(action) \
         do { ; } while(0)
#else
# define ON_GL_ERROR(action)                       \
         do                                        \
           {                                       \
           const char *__glErr=CoreGl_lastError(); \
           ON_ERROR(__glErr,__glErr,action);       \
           } while(0)
#endif

struct _CoreGlManager
{
bool initDone;
bool coreProfile;
bool embedded;
int majorVersion,minorVersion;
int shadingLanguageVersion;
int extensionCount;
char ** extensions;
bool hasTessControlShaders;
bool hasTessEvaluationShaders;
bool hasGeometryShaders;
bool hasComputeShaders;'''%('-'*76))

for e in extensions:
  header.write(r'''
bool has_%s;'''%e.name)

header.write(r'''
};

extern TRANSGL_API_VAR struct _CoreGlManager _CoreGl_mgr;

TRANSGL_API
bool // success
CoreGl_init(void);

TRANSPROG_INLINE
bool // OpenGL core profile is supported
CoreGl_coreProfile(void)
{
return _CoreGl_mgr.coreProfile;
}

TRANSPROG_INLINE
bool // OpenGL is actually OpenGL ES
CoreGl_embedded(void)
{
return _CoreGl_mgr.embedded;
}

TRANSPROG_INLINE
int // OpenGL major version
CoreGl_majorVersion(void)
{
return _CoreGl_mgr.majorVersion;
}

TRANSPROG_INLINE
int // OpenGL minor version
CoreGl_minorVersion(void)
{
return _CoreGl_mgr.minorVersion;
}

TRANSPROG_INLINE
bool // success
CoreGl_isAtLeastVersion(int major,
                        int minor)
{
return (_CoreGl_mgr.majorVersion>major)||
       ((_CoreGl_mgr.majorVersion==major)&&
        (_CoreGl_mgr.minorVersion>=minor));
}

TRANSPROG_INLINE
int // GLSL version
CoreGl_shadingLanguageVersion(void)
{
return _CoreGl_mgr.shadingLanguageVersion;
}

TRANSPROG_INLINE
bool // tessellation evaluation shaders are available
CoreGl_hasTessEvaluationShaders(void)
{
return _CoreGl_mgr.hasTessEvaluationShaders;
}

TRANSPROG_INLINE
bool // tessellation control shaders are available
CoreGl_hasTessControlShaders(void)
{
return _CoreGl_mgr.hasTessControlShaders;
}

TRANSPROG_INLINE
bool // geometry shaders are available
CoreGl_hasGeometryShaders(void)
{
return _CoreGl_mgr.hasGeometryShaders;
}

TRANSPROG_INLINE
bool // compute shaders are available
CoreGl_hasComputeShaders(void)
{
return _CoreGl_mgr.hasComputeShaders;
}

TRANSPROG_INLINE
const char * const * // extension name array
CoreGl_extensions(int *out_extensionCount)
{
*out_extensionCount=_CoreGl_mgr.extensionCount;
return (const char * const *)_CoreGl_mgr.extensions;
}

TRANSGL_API
bool // success
CoreGl_hasExtension(const char *extensionName);

TRANSGL_API
void * // symbol or 0
CoreGl_extensionSymbol(const char *symbolName);

#define CoreGl_extensionFunction(out_function,fnctName)                \
        do                                                             \
          {                                                            \
          *((void **)(out_function))=CoreGl_extensionSymbol(fnctName); \
          } while(0)

TRANSGL_API
void * // symbol or 0
CoreGl_extensionSymbolWithSuffix(const char *symbolName,
                                 const char * const *suffixList);

#define CoreGl_extensionFunctionWithSuffix(out_function,fnctName,suffixList) \
        do                                                                   \
          {                                                                  \
          *((void **)(out_function))=                                        \
            CoreGl_extensionSymbolWithSuffix(fnctName,suffixList);           \
          } while(0)

TRANSGL_API
const char * // last OpenGL error message if any
CoreGl_lastError(void);

TRANSGL_API
char * // allocated compilation/link error/warning message if any
CoreGl_shaderStatus(GLuint shader,
                    bool *out_error);

''')

source.write(r'''
#if defined _MSC_VER
# define _CRT_SECURE_NO_WARNINGS
# define snprintf _snprintf
#endif

#include "TransGl/coreGl.h"
//
// The preceding inclusion can be changed in order to find codeGl.h in its
// actual location (independant from the TransProg library).
//
#if !defined _WIN32
# include <dlfcn.h>
#endif

#if defined TRANSGL_API_H
  #include "TransProg/memory.h"
  const char * _CoreGl_querySpecificExtensions(void); // defined elsewhere
#else // TransProg library is not used
  static
  const char * _CoreGl_querySpecificExtensions(void)
  { return (const char *)0; } // fake definition
#endif

#if defined _WIN32
# define GL_LIBRARY_NAME  "OPENGL32.DLL"
#elif defined __APPLE__
# define GL_LIBRARY_NAME  "/System/Library/Frameworks/OpenGL.framework" \
                          "/Libraries/libGL.dylib"
#elif defined __ANDROID__
# define GL_LIBRARY_NAME  "libGLESv2.so"
  TRANSPROG_EXTERN_C void * eglGetProcAddress(const char *);
#else
# define GL_LIBRARY_NAME  "libGL.so"
  TRANSPROG_EXTERN_C void * glXGetProcAddress(const char *);
#endif

TRANSGL_API_VAR struct _CoreGlManager _CoreGl_mgr;

static
void
CoreGl_addExtensionList(const char *extensionList)
{
if(extensionList)
  {
  for(int run=0;run<2;++run)
    {
    const char *beginList=extensionList;
    const char *endList=beginList+strlen(beginList);
    int extCount=_CoreGl_mgr.extensionCount;
    while(beginList<endList)
      {
      int extLen=(int)strcspn(beginList," ");
      if(extLen)
        {
        if(run!=0)
          {
          char *ext=Memory_alloc(extLen+1,char);
          memcpy(ext,beginList,extLen);
          ext[extLen]='\0';
          _CoreGl_mgr.extensions[extCount]=ext;
          }
        ++extCount;
        }
      beginList+=extLen+1;
      }
    if(run==0)
      {
      _CoreGl_mgr.extensions=Memory_reAlloc(_CoreGl_mgr.extensions,
                                            extCount+1,char *);
      }
    else
      {
      _CoreGl_mgr.extensions[extCount]=(char *)0;
      _CoreGl_mgr.extensionCount=extCount;
      }
    }
  }
}
''')

for e in extensions:
  source.write(r'''
static
bool // extension is available
CoreGl_load_%s(void)
{
bool hasExt='''%e.name)
  if e.name.startswith('VERSION_'):
    v=e.name.split('_')
    source.write(r'''CoreGl_isAtLeastVersion(%d,%d);
'''%(int(v[1]),int(v[2])))
  else:
    source.write(r'''CoreGl_hasExtension("GL_%s");
'''%e.name)
  for a in e.apiEntries:
    source.write(r'''CoreGl_extensionFunction(&CoreGl_%s,
                         "gl%s");
'''%(a.name,a.name))
    if e.name.startswith('VERSION_'):
      source.write(r'''if(!CoreGl_%s)
  {
  hasExt=false;
  }
'''%a.name)
    else:
      source.write(r'''ON_ERROR(hasExt&&!CoreGl_%s,
         "missing gl%s() in GL_%s extension\n",
         {});
'''%(a.name,a.name,e.name))
  source.write(r'''return hasExt;
}
''')

source.write(r'''
bool // success
CoreGl_init(void)
{
if(!_CoreGl_mgr.initDone)
  {
  _CoreGl_mgr.coreProfile=false;
  _CoreGl_mgr.embedded=false;
  _CoreGl_mgr.majorVersion=0;
  _CoreGl_mgr.minorVersion=0;
  _CoreGl_mgr.shadingLanguageVersion=0;
  _CoreGl_mgr.extensionCount=0;
  _CoreGl_mgr.extensions=Memory_alloc(1,char *);
  _CoreGl_mgr.extensions[0]=(char *)0;
  _CoreGl_mgr.hasTessControlShaders=false;
  _CoreGl_mgr.hasTessEvaluationShaders=false;
  _CoreGl_mgr.hasGeometryShaders=false;
  _CoreGl_mgr.hasComputeShaders=false;
  CoreGl_extensionFunction(&glGetError,"glGetError");
  ON_ERROR(!glGetError,"glGetError() not found",{ return false; });
  CoreGl_extensionFunction(&glGetIntegerv,"glGetIntegerv");
  ON_ERROR(!glGetIntegerv,"glGetIntegerv() not found",{ return false; });
  CoreGl_extensionFunction(&glGetString,"glGetString");
  ON_ERROR(!glGetString,"glGetString() not found",{ return false; });
  int value=-1;
  glGetIntegerv(GL_CONTEXT_PROFILE_MASK,&value); // may fail
  if(value==-1) glGetError(); // clear previous error
  else
    {
    if(value&GL_CONTEXT_CORE_PROFILE_BIT)
      {
      _CoreGl_mgr.coreProfile=true;
      }
    }
  bool success=false;
  const char *version=(const char *)glGetString(GL_VERSION);
  ON_GL_ERROR({ return false; });
  if(version)
    {
    int major,minor;
    if(sscanf(version," %d.%d ",&major,&minor)==2)
      {
      _CoreGl_mgr.majorVersion=major;
      _CoreGl_mgr.minorVersion=minor;
      success=true;
      }
    else if(sscanf(version," OpenGL ES %d.%d ",&major,&minor)==2)
      {
      _CoreGl_mgr.embedded=true;
      _CoreGl_mgr.majorVersion=major;
      _CoreGl_mgr.minorVersion=minor;
      success=true;
      }
    }
  ON_ERROR(!success,"OpenGL version not detected",{ return false; });
  success=false;
  const char *glslVersion=(const char *)
                          glGetString(GL_SHADING_LANGUAGE_VERSION);
  ON_GL_ERROR({ return false; });
  if(glslVersion)
    {
    int major,minor;
    if(((sscanf(glslVersion," %d.%d ",&major,&minor)==2)||
        (sscanf(glslVersion," OpenGL ES %d.%d ",&major,&minor)==2))&&
       (minor<100))
      {
      _CoreGl_mgr.shadingLanguageVersion=100*major+minor;
      success=true;
      }
    }
  ON_ERROR(!success,"OpenGL GLSL version not detected",{ return false; });
  CoreGl_extensionFunction(&glGetStringi,"glGetStringi"); // may fail
  if(glGetStringi) // modern means to detect extensions
    {
    int extCount=0;
    glGetIntegerv(GL_NUM_EXTENSIONS,&extCount); // may fail
    if(!extCount) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.extensions=Memory_reAlloc(_CoreGl_mgr.extensions,
                                            extCount+1,char *);
      for(int i=0;i<extCount;++i)
        {
        const char *s=(const char *)glGetStringi(GL_EXTENSIONS,i); // may fail
        if(!s) glGetError(); // clear previous error
        else
          {
          int extLen=(int)strlen(s);
          char *ext=Memory_alloc(extLen+1,char);
          memcpy(ext,s,extLen+1);
          _CoreGl_mgr.extensions[_CoreGl_mgr.extensionCount++]=ext;
          }
        }
      _CoreGl_mgr.extensions[_CoreGl_mgr.extensionCount]=(char *)0;
      }
    }
  if(!_CoreGl_mgr.extensionCount) // fallback to legacy extension detection
    {
    const char *extList=(const char *)glGetString(GL_EXTENSIONS); // may fail
    if(!extList) glGetError(); // clear previous error
    else
      {
      CoreGl_addExtensionList(extList);
      }
    }
  CoreGl_addExtensionList(_CoreGl_querySpecificExtensions());
''')

for e in extensions:
  source.write(r'''  _CoreGl_mgr.has_%s=CoreGl_load_%s();
'''%(e.name,e.name))

source.write(r'''  if(glCreateShader&&glDeleteShader)
    {
    int featureTest;
    featureTest=glCreateShader(GL_TESS_CONTROL_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasTessControlShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    featureTest=glCreateShader(GL_TESS_EVALUATION_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasTessEvaluationShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    featureTest=glCreateShader(GL_GEOMETRY_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasGeometryShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    featureTest=glCreateShader(GL_COMPUTE_SHADER); // may fail
    if(!featureTest) glGetError(); // clear previous error
    else
      {
      _CoreGl_mgr.hasComputeShaders=true;
      glDeleteShader(featureTest);
      ON_GL_ERROR({ return false; });
      }
    }
  _CoreGl_mgr.initDone=true;
  }
return true;
}

bool // success
CoreGl_hasExtension(const char *extensionName)
{
int count;
const char * const *ext=CoreGl_extensions(&count);
for(int i=0;i<count;++i)
  {
  if(!strcmp(ext[i],extensionName))
    {
    return true;
    }
  }
return false;
}

void * // symbol or 0
CoreGl_extensionSymbol(const char *symbolName)
{
void *symbol=(void *)0;
#if defined _WIN32
  PROC procAddr=wglGetProcAddress(symbolName);
  symbol=*((void **)&procAddr);
#elif defined __APPLE__
  // FIXME: no ???GetProcAddress() utility
#elif defined __ANDROID__
  // FIXME: eglGetProcAddress() only works for extension functions
  // symbol=eglGetProcAddress(symbolName);
#else 
  symbol=glXGetProcAddress(symbolName);
#endif
if(!symbol) // not found with a specific utility... may be the application
  {         // is linked against GL...
  static void *currentBinary=(void *)0;
  if(!currentBinary)
    {
#if defined _WIN32
    currentBinary=(void *)GetModuleHandle((LPCTSTR)0);
#else
    currentBinary=dlopen((const char *)0,RTLD_LAZY);
#endif
    }
  if(currentBinary)
    {
#if defined _WIN32
    FARPROC procAddr=GetProcAddress((HMODULE)currentBinary,symbolName);
    symbol=*((void **)&procAddr);
#else
    symbol=dlsym(currentBinary,symbolName);
#endif
    }
  }
if(!symbol) // not found in the current binary... may be the application
  {         // is a dynamic library which is linked against GL...
  static void *glLibrary=(void *)0;
  if(!glLibrary)
    {
#if defined _WIN32
    glLibrary=(void *)LoadLibrary(GL_LIBRARY_NAME);
#else
    glLibrary=dlopen(GL_LIBRARY_NAME,RTLD_LAZY);
#endif
    }
  if(glLibrary)
    {
#if defined _WIN32
    FARPROC procAddr=GetProcAddress((HMODULE)glLibrary,symbolName);
    symbol=*((void **)&procAddr);
#else
    symbol=dlsym(glLibrary,symbolName);
#endif
    }
  }
return symbol;
}

void * // symbol or 0
CoreGl_extensionSymbolWithSuffix(const char *symbolName,
                                 const char * const *suffixList)
{
for(int i=0;suffixList[i];++i)
  {
  char fullName[0x100];
  snprintf(fullName,0x100,"%s%s",symbolName,suffixList[i]);
  fullName[0xFF]='\0';
  void *symbol=CoreGl_extensionSymbol(fullName);
  if(symbol)
    {
    return symbol;
    }
  }
return (void *)0;
}

const char * // last OpenGL error message if any
CoreGl_lastError(void)
{
int err=glGetError();
switch(err)
  {
  case GL_NO_ERROR:
    {
    return (const char *)0;
    }
  case GL_INVALID_ENUM:
    {
    return "GL_INVALID_ENUM";
    }
  case GL_INVALID_VALUE:
    {
    return "GL_INVALID_VALUE";
    }
  case GL_INVALID_OPERATION:
    {
    return "GL_INVALID_OPERATION";
    }
  case GL_STACK_OVERFLOW:
    {
    return "GL_STACK_OVERFLOW";
    }
  case GL_STACK_UNDERFLOW:
    {
    return "GL_STACK_UNDERFLOW";
    }
  case GL_OUT_OF_MEMORY:
    {
    return "GL_OUT_OF_MEMORY";
    }
  case GL_INVALID_FRAMEBUFFER_OPERATION:
    {
    return "GL_INVALID_FRAMEBUFFER_OPERATION";
    }
  default:
    {
    static char buffer[0x20];
    sprintf(buffer,"unknown OpenGL error 0x%.4X",err);
    return buffer;
    }
  }
}

char * // allocated compilation/link error/warning message if any
CoreGl_shaderStatus(GLuint shader,
                    bool *out_error)
{
bool isProgram;
GLint status=GL_FALSE;
GLint len=0;
if(glIsProgram(shader))
  {
  isProgram=true;
  glGetProgramiv(shader,GL_LINK_STATUS,&status);
  ON_GL_ERROR({});
  glGetProgramiv(shader,GL_INFO_LOG_LENGTH,&len);
  ON_GL_ERROR({});
  }
else
  {
  isProgram=false;
  glGetShaderiv(shader,GL_COMPILE_STATUS,&status);
  ON_GL_ERROR({});
  glGetShaderiv(shader,GL_INFO_LOG_LENGTH,&len);
  ON_GL_ERROR({});
  }
char *log=(char *)0;
*out_error=!status;
if(len>1)
  {
  log=Memory_initAlloc(len,char);
  GLint realLen=0;
  if(isProgram)
    {
    glGetProgramInfoLog(shader,len,&realLen,log);
    ON_GL_ERROR({ Memory_free(log); log=(char *)0; });
    }
  else
    {
    glGetShaderInfoLog(shader,len,&realLen,log);
    ON_GL_ERROR({ Memory_free(log); log=(char *)0; });
    }
  if(log)
    {
    bool empty=true;
    for(int i=0;(i<len)&&log[i];++i)
      {
      if(!isspace(log[i]))
        {
        empty=false;
        break;
        }
      }
    if(empty)
      {
      Memory_free(log);
      log=(char *)0;
      }
    }
  }
if(!status&&!log)
  {
  log=Memory_alloc(64,char);
  if(isProgram)
    {
    sprintf(log,"shader program link failure\n");
    }
  else
    {
    sprintf(log,"shader compilation failure\n");
    }
  }
return log;
}

''')

for e in extensions:
  header.write(r'''//---- %s ----

TRANSPROG_INLINE
bool
CoreGl_has_%s(void)
{
return _CoreGl_mgr.has_%s;
}

'''%(e.name,e.name,e.name))
  source.write(r'''//---- %s ----

'''%e.name)
  for a in e.apiEntries:
    header.write(r'''extern TRANSGL_API_VAR
PFNGL%sPROC
CoreGl_%s;
#define gl%s \
        CoreGl_%s

'''%(a.typeName,a.name,a.name,a.name))
    source.write(r'''TRANSGL_API_VAR
PFNGL%sPROC
CoreGl_%s;

'''%(a.typeName,a.name))

header.write(r'''#endif // TRANSGL_COREGL_H

//'''+'-'*76+'\n')
source.write('//'+'-'*76+'\n')
header.close()
source.close()

#-----------------------------------------------------------------------------
