#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import sys ; sys.path.insert(0,'../../lib')
import TransProgConfig

import glob

#-------- Create the construction tool ---------------------------------------

tool=TransProgConfig.Tool()

tool.settings['DEFINES']['WIN32_TRANSGL_DLL']=None

if tool.sysName().startswith('Darwin'):
  tool.settings['LDFLAGS']+=['-framework','OpenGL',
                             '-framework','AppKit',
                             '-framework','Carbon']
  import platform
  if platform.release()=='9.7.0': # MacOSX 10.5 bug!
    glLib='/System/Library/Frameworks/OpenGL.framework/'+ \
          'Versions/A/Libraries/libGL.dylib'
    tool.settings['LDFLAGS']+=['-Wl,-dylib_file,%s:%s' % (glLib,glLib)]
  try:
    tool.settings['CFLAGS'].remove('-xc') # allow objective-c source
  except:
    pass
elif tool.sysName().startswith('Windows'):
  tool.settings['LIBS']+=['OPENGL32','GDI32'] # GLU32 is not needed anymore
else:
  tool.settings['INCDIRS']+=['/usr/X11R7/include','/usr/X11R6/include']
  tool.settings['LIBDIRS']+=['/usr/X11R7/lib','/usr/X11R6/lib']
  tool.settings['LIBS']+=['GL','X11'] # GLU is not needed anymore

topDir=tool.settings['TRANSPROG_DIR']

#-------- Enumerate targets and source files ---------------------------------

headers=[f for f in glob.glob(topDir+'/include/TransGl/*.h')
         if not f.endswith('transGlAsyncCalls.h')]

tool.Template(topDir+'/include/TransGl/transGlAsyncCalls.h',
              [topDir+'/lib/asyncCallGenerator.py']+headers,
              ['-h',topDir+'/include/TransProg/platform.h',
               '-h',topDir+'/include/TransGl/api.h'])

if   tool.sysName().startswith('Darwin'):  glView='glView_cocoa.m'
elif tool.sysName().startswith('Windows'): glView='glView_win32.c'
else:                                      glView='glView_x11.c'

tool.Library(topDir+'/lib/TransGl',
             [glView,
              'coreGl.c',
              'glDraw.c',
              'viewer.c',
              'vrRoom.c'])

tool.Clean('swig_*')

tool.run()

#-----------------------------------------------------------------------------
