//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/viewer.h"

static
bool // keep view
Viewer_callback(GlView *view)
{
const GfxState *state=GlView_state(view);
// GfxState_displayEvent(state);
Viewer *v=(Viewer *)GlView_data(view);
if(!v)
  {
  return false; // destroyed
  }
bool useCallback=true;
if(Examiner_handleEvent(&v->examiner,state,&v->viewPoint,v->refDistance))
  {
  GlView_postRedraw(view);
  }
switch(GfxState_event(state))
  {
  case GFX_CONFIG_EVT:
    {
    if(GfxState_visible(state))
      {
      real64_t ratio=GfxState_width(state)/(real64_t)GfxState_height(state);
      if(GlView_stereoMode()==GFX_SPLIT_STEREO)
        {
        ratio*=0.5;
        }
      Perspective_setAspectRatio(&v->perspective,ratio);
      GlView_postRedraw(view);
      }
    break;
    }
  case GFX_DRAW_EVT:
    {
    if(GfxState_visible(state))
      {
      switch(GlView_stereoMode())
        {
        case GFX_NO_STEREO:
          {
          glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
          Gfx_computeProjectionMatrix(v->leftProjMtx,&v->perspective);
          Gfx_computeViewMatrix(v->leftViewMtx,&v->viewPoint);
          v->viewport.x=0;
          v->viewport.y=0;
          v->viewport.w=GfxState_width(state);
          v->viewport.h=GfxState_height(state);
          glViewport(v->viewport.x,v->viewport.y,
                     v->viewport.w,v->viewport.h);
          break;
          }
        case GFX_ACTIVE_STEREO:
          {
          Gfx_computeStereoProjectionMatrices(v->leftProjMtx,
                                              v->rightProjMtx,
                                              &v->perspective);
          Gfx_computeStereoViewMatrices(v->leftViewMtx,
                                        v->rightViewMtx,
                                        &v->viewPoint,
                                        &v->perspective);
          glDrawBuffer(GL_BACK_RIGHT);
          glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
          v->viewport.x=0;
          v->viewport.y=0;
          v->viewport.w=GfxState_width(state);
          v->viewport.h=GfxState_height(state);
          glViewport(v->viewport.x,v->viewport.y,
                     v->viewport.w,v->viewport.h);
          v->right=true;
          (*v->callback)(v);
          if(!GlView_data(view))
            {
            return false; // destroyed during callback
            }
          v->right=false;
          glDrawBuffer(GL_BACK_LEFT);
          glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
          break;
          }
        case GFX_SPLIT_STEREO:
          {
          glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
          Gfx_computeStereoProjectionMatrices(v->leftProjMtx,
                                              v->rightProjMtx,
                                              &v->perspective);
          Gfx_computeStereoViewMatrices(v->leftViewMtx,
                                        v->rightViewMtx,
                                        &v->viewPoint,
                                        &v->perspective);
          int halfWidth=GfxState_width(state)>>1;
          v->viewport.x=halfWidth;
          v->viewport.y=0;
          v->viewport.w=halfWidth;
          v->viewport.h=GfxState_height(state);
          glViewport(v->viewport.x,v->viewport.y,
                     v->viewport.w,v->viewport.h);
          v->right=true;
          (*v->callback)(v);
          if(!GlView_data(view))
            {
            return false; // destroyed during callback
            }
          v->right=false;
          v->viewport.x=0;
          v->viewport.y=0;
          v->viewport.w=halfWidth;
          v->viewport.h=GfxState_height(state);
          glViewport(v->viewport.x,v->viewport.y,
                     v->viewport.w,v->viewport.h);
          break;
          }
        }
      }
    else
      {
      useCallback=false;
      }
    break;
    }
  default:
    {
    break;
    }
  }
if(useCallback)
  {
  (*v->callback)(v);
  }
return GlView_data(view) ? true : false; // not destroyed during callback
}

void
Viewer_init(Viewer *v,
            int initX,
            int initY,
            int initWidth,
            int initHeight,
            bool borderless,
            ViewerCallback callback,
            void *data)
{
v->callback=callback;
v->data=data;
SpatialBasis_init(&v->viewPoint);
Perspective_init(&v->perspective);
Examiner_init(&v->examiner);
const Frustum *frustum=Perspective_frustum(&v->perspective);
v->refDistance=0.5*(frustum->nearDistance+frustum->farDistance);
v->right=false;
v->view=GlView_open(initX,initY,initWidth,initHeight,borderless,
                    &Viewer_callback,v);
memset(v->leftProjMtx,0,sizeof(v->leftProjMtx));
memset(v->rightProjMtx,0,sizeof(v->rightProjMtx));
memset(v->leftViewMtx,0,sizeof(v->leftViewMtx));
memset(v->rightViewMtx,0,sizeof(v->rightViewMtx));
v->viewport.x=0;
v->viewport.y=0;
v->viewport.w=0;
v->viewport.h=0;
//-- change initial values of somme common settings --
// FIXME: these settings are global, not local to a specific viewer
glClearColor((real32_t)(0x60/255.0),
             (real32_t)(0x60/255.0),
             (real32_t)(0x60/255.0),1.0f);        // initial is 0,0,0,0
glEnable(GL_DEPTH_TEST);                          // initial is disabled
glEnable(GL_CULL_FACE);                           // initial is disabled
glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); // initial is 1,0
glPixelStorei(GL_PACK_ALIGNMENT,1);               // initial is 4
glPixelStorei(GL_UNPACK_ALIGNMENT,1);             // initial is 4
}

void
Viewer_destroy(Viewer *v)
{
v->view->data=(void *)0; // mark as destroyed
GlView_postRedraw(v->view); // force callback invocation
SpatialBasis_destroy(&v->viewPoint);
Perspective_destroy(&v->perspective);
Examiner_destroy(&v->examiner);
}

void
Viewer_setRefDistanceFromPoint(Viewer *v,
                               real64_t px,
                               real64_t py,
                               real64_t pz)
{
SpatialBasis_globalToLocalPosition(&v->viewPoint,&px,&py,&pz);
const Frustum *frustum=Perspective_frustum(&v->perspective);
if(px<frustum->nearDistance)
  {
  px=0.5*(frustum->nearDistance+frustum->farDistance);
  }
Viewer_setRefDistance(v,px);
}

//----------------------------------------------------------------------------
