//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/examiner.h"

bool // event used
Examiner_handleEvent(Examiner *e,
                     const GfxState *state,
                     SpatialBasis *location,
                     real64_t refDistance)
{
switch(GfxState_event(state))
  {
  case GFX_BUTTON_EVT:
    {
    if(GfxState_pressed(state))
      {
      switch(GfxState_button(state))
        {
        case GFX_LEFT_BTN:
        case GFX_MIDDLE_BTN:
        case GFX_RIGHT_BTN:
          {
          e->pending=true;
          e->button=GfxState_button(state);
          e->mouseX=GfxState_mouseX(state);
          e->mouseY=GfxState_mouseY(state);
          return true;
          }
        default:
          {
          break;
          }
        }
      }
    else
      {
      if(e->pending&&(GfxState_button(state)==e->button))
        {
        e->pending=false;
        return true;
        }
      }
    break;
    }
  case GFX_MOTION_EVT:
    {
    if(e->pending)
      {
      int w=GfxState_width(state);
      int h=GfxState_height(state);
      int oldX=e->mouseX;
      int oldY=e->mouseY;
      e->mouseX=GfxState_mouseX(state);
      e->mouseY=GfxState_mouseY(state);
      real64_t dx=2.0*((real64_t)(e->mouseX-oldX))/w;
      real64_t dy=2.0*((real64_t)(e->mouseY-oldY))/h;
      switch(e->button)
        {
        case GFX_LEFT_BTN:
          {
          real64_t lastDx=0.0;
          real64_t lastDy=dy;
          real64_t lastDz=-dx;
          real64_t r=sqrt(lastDy*lastDy+lastDz*lastDz);
          if(r)
            {
            lastDy/=r;
            lastDz/=r;
            }
          real64_t lastAngle=r*M_PI;
          SpatialBasis_translate(location,refDistance,0.0,0.0);
          SpatialBasis_rotate(location,lastDx,lastDy,lastDz,lastAngle);
          SpatialBasis_translate(location,-refDistance,0.0,0.0);
          return true;
          }
        case GFX_MIDDLE_BTN:
          {
          SpatialBasis_translate(location,
                                 0.0,
                                 0.5*refDistance*dx,
                                 0.5*refDistance*dy);
          return true;
          }
        case GFX_RIGHT_BTN:
          {
          SpatialBasis_translate(location,refDistance*dy,0.0,0.0);
          return true;
          }
        default:
          {
          break;
          }
        }
      }
    break;
    }
  default:
    {
    break;
    }
  }
return false;
}

//----------------------------------------------------------------------------
