#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

content='//'+'-'*76+r'''
// Copyright (C) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#line 42 THIS_FILE

static
void
Skeleton_applyHint_empty(const real64_t *values,
                         SpatialBasis *inout_basis)
{
(void)values;
(void)inout_basis;
}
'''

import sys
import os

if len(sys.argv)<2:
  sys.stderr.write(
             'usage: %s file_name\n'
             % sys.argv[0])
  sys.exit(1)

fileName=sys.argv[1]

hintList=[]

def generateHints(pool,l,h):
  if not l:
    count=0
    hint=0
    d={ 'T':0, 'X':1, 'Y':2, 'Z':3 }
    for i in h:
      i=d.get(i)
      hint|=(i<<(2*count))
      count+=1
    hint|=(count<<8)
    hintList.append((h,'0x%8.8x'%hint))
  else:
    for i in pool:
      if not i in h:
        generateHints(pool,l-1,h+i)

for l in range(4):
  h=''
  generateHints('TXYZ',l+1,h)

for h in hintList:
  content+='''
static
void
Skeleton_applyHint_%s(const real64_t *values,
                   %s SpatialBasis *inout_basis)
{'''%(h[0],' '*len(h[0]))
  for t in h[0]:
    if t=='T':
      content+='''
SpatialBasis_translate(inout_basis,values[0],values[1],values[2]);'''
    elif t=='X':
      content+='''
SpatialBasis_roll(inout_basis,values[3]);'''
    elif t=='Y':
      content+='''
SpatialBasis_pitch(inout_basis,values[4]);'''
    elif t=='Z':
      content+='''
SpatialBasis_yaw(inout_basis,values[5]);'''
  content+='''
}
'''

content+='''
static
void (*
Skeleton_chooseApplyHint(int hint)
)(const real64_t *,SpatialBasis *)
{
switch(hint&0x00000FFF)
  {
  case 0x00000000: return &Skeleton_applyHint_empty;'''
for h in hintList:
  content+='''
  case %s: return &Skeleton_applyHint_%s;'''%(h[1],h[0])
content+='''
  }
return (void (*)(const real64_t *,SpatialBasis *))0;
}

//----------------------------------------------------------------------------
'''

thisFile='"%s"' % os.path.normpath(sys.argv[0]).replace(os.path.sep,'/')

open(fileName,'w').write(content.replace('THIS_FILE',thisFile))

