//----------------------------------------------------------------------------
// Copyright (c) 2008-2015 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/surface.h"

#include "TransProg/memory.h"
#include "TransProg/math.h"
#include "TransProg/path.h"

#define ADD_V(vx,vy,vz)                                      \
        if(vertices)                                         \
          {                                                  \
          switch(sizeOfVertex)                               \
            {                                                \
            case 2:                                          \
              {                                              \
              uint16_t *ptr=(uint16_t *)vertices;            \
              ptr[0]=real64ToHalf((vx));                     \
              ptr[1]=real64ToHalf((vy));                     \
              ptr[2]=real64ToHalf((vz));                     \
              vertices=ptr+3;                                \
              break;                                         \
              }                                              \
            case 4:                                          \
              {                                              \
              real32_t *ptr=(real32_t *)vertices;            \
              ptr[0]=(real32_t)(vx);                         \
              ptr[1]=(real32_t)(vy);                         \
              ptr[2]=(real32_t)(vz);                         \
              vertices=ptr+3;                                \
              break;                                         \
              }                                              \
            case 8:                                          \
              {                                              \
              real64_t *ptr=(real64_t *)vertices;            \
              ptr[0]=(real64_t)(vx);                         \
              ptr[1]=(real64_t)(vy);                         \
              ptr[2]=(real64_t)(vz);                         \
              vertices=ptr+3;                                \
              break;                                         \
              }                                              \
            default:                                         \
              {                                              \
              DROP_DEAD("bad value for sizeOfVertex");       \
              }                                              \
            }                                                \
          }

#define ADD_N(nx,ny,nz)                                      \
        if(normals)                                          \
          {                                                  \
          switch(sizeOfNormal)                               \
            {                                                \
            case 2:                                          \
              {                                              \
              uint16_t *ptr=(uint16_t *)normals;             \
              ptr[0]=real64ToHalf((nx));                     \
              ptr[1]=real64ToHalf((ny));                     \
              ptr[2]=real64ToHalf((nz));                     \
              normals=ptr+3;                                 \
              break;                                         \
              }                                              \
            case 4:                                          \
              {                                              \
              real32_t *ptr=(real32_t *)normals;             \
              ptr[0]=(real32_t)(nx);                         \
              ptr[1]=(real32_t)(ny);                         \
              ptr[2]=(real32_t)(nz);                         \
              normals=ptr+3;                                 \
              break;                                         \
              }                                              \
            case 8:                                          \
              {                                              \
              real64_t *ptr=(real64_t *)normals;             \
              ptr[0]=(real64_t)(nx);                         \
              ptr[1]=(real64_t)(ny);                         \
              ptr[2]=(real64_t)(nz);                         \
              normals=ptr+3;                                 \
              break;                                         \
              }                                              \
            default:                                         \
              {                                              \
              DROP_DEAD("bad value for sizeOfNormal");       \
              }                                              \
            }                                                \
          }

#define ADD_T(tx,ty)                                         \
        if(texCoords)                                        \
          {                                                  \
          switch(sizeOfTexCoord)                             \
            {                                                \
            case 2:                                          \
              {                                              \
              uint16_t *ptr=(uint16_t *)texCoords;           \
              ptr[0]=real64ToHalf((tx));                     \
              ptr[1]=real64ToHalf((ty));                     \
              texCoords=ptr+2;                               \
              break;                                         \
              }                                              \
            case 4:                                          \
              {                                              \
              real32_t *ptr=(real32_t *)texCoords;           \
              ptr[0]=(real32_t)(tx);                         \
              ptr[1]=(real32_t)(ty);                         \
              texCoords=ptr+2;                               \
              break;                                         \
              }                                              \
            case 8:                                          \
              {                                              \
              real64_t *ptr=(real64_t *)texCoords;           \
              ptr[0]=(real64_t)(tx);                         \
              ptr[1]=(real64_t)(ty);                         \
              texCoords=ptr+2;                               \
              break;                                         \
              }                                              \
            default:                                         \
              {                                              \
              DROP_DEAD("bad value for sizeOfTexCoord");     \
              }                                              \
            }                                                \
          }

#define ADD_TRI(v0,v1,v2)                                    \
        if(indices)                                          \
          {                                                  \
          switch(sizeOfIndex)                                \
            {                                                \
            case 1:                                          \
              {                                              \
              uint8_t *ptr=(uint8_t *)indices;               \
              ptr[0]=(uint8_t)(v0);                          \
              ptr[1]=(uint8_t)(v1);                          \
              ptr[2]=(uint8_t)(v2);                          \
              indices=ptr+3;                                 \
              break;                                         \
              }                                              \
            case 2:                                          \
              {                                              \
              uint16_t *ptr=(uint16_t *)indices;             \
              ptr[0]=(uint16_t)(v0);                         \
              ptr[1]=(uint16_t)(v1);                         \
              ptr[2]=(uint16_t)(v2);                         \
              indices=ptr+3;                                 \
              break;                                         \
              }                                              \
            case 4:                                          \
              {                                              \
              uint32_t *ptr=(uint32_t *)indices;             \
              ptr[0]=(uint32_t)(v0);                         \
              ptr[1]=(uint32_t)(v1);                         \
              ptr[2]=(uint32_t)(v2);                         \
              indices=ptr+3;                                 \
              break;                                         \
              }                                              \
            case 8:                                          \
              {                                              \
              uint64_t *ptr=(uint64_t *)indices;             \
              ptr[0]=(uint64_t)(v0);                         \
              ptr[1]=(uint64_t)(v1);                         \
              ptr[2]=(uint64_t)(v2);                         \
              indices=ptr+3;                                 \
              break;                                         \
              }                                              \
            default:                                         \
              {                                              \
              DROP_DEAD("bad value for sizeOfIndex");        \
              }                                              \
            }                                                \
          }

#define GET_V(vx,vy,vz,index)                                \
        switch(sizeOfVertex)                                 \
          {                                                  \
          case 2:                                            \
            {                                                \
            const uint16_t *ptr=(const uint16_t *)vertices;  \
            ptr+=3*(index);                                  \
            (vx)=halfToReal64(ptr[0]);                       \
            (vy)=halfToReal64(ptr[1]);                       \
            (vz)=halfToReal64(ptr[2]);                       \
            break;                                           \
            }                                                \
          case 4:                                            \
            {                                                \
            const real32_t *ptr=(const real32_t *)vertices;  \
            ptr+=3*(index);                                  \
            (vx)=(real64_t)ptr[0];                           \
            (vy)=(real64_t)ptr[1];                           \
            (vz)=(real64_t)ptr[2];                           \
            break;                                           \
            }                                                \
          case 8:                                            \
            {                                                \
            const real64_t *ptr=(const real64_t *)vertices;  \
            ptr+=3*(index);                                  \
            (vx)=(real64_t)ptr[0];                           \
            (vy)=(real64_t)ptr[1];                           \
            (vz)=(real64_t)ptr[2];                           \
            break;                                           \
            }                                                \
          default:                                           \
            {                                                \
            (vx)=(vy)=(vz)=0.0;                              \
            DROP_DEAD("bad value for sizeOfVertex");         \
            }                                                \
          }

#define GET_TRI(v0,v1,v2,index)                              \
        switch(sizeOfIndex)                                  \
          {                                                  \
          case 1:                                            \
            {                                                \
            const uint8_t *ptr=(const uint8_t *)indices;     \
            ptr+=3*(index);                                  \
            (v0)=(int)ptr[0];                                \
            (v1)=(int)ptr[1];                                \
            (v2)=(int)ptr[2];                                \
            break;                                           \
            }                                                \
          case 2:                                            \
            {                                                \
            const uint16_t *ptr=(const uint16_t *)indices;   \
            ptr+=3*(index);                                  \
            (v0)=(int)ptr[0];                                \
            (v1)=(int)ptr[1];                                \
            (v2)=(int)ptr[2];                                \
            break;                                           \
            }                                                \
          case 4:                                            \
            {                                                \
            const uint32_t *ptr=(const uint32_t *)indices;   \
            ptr+=3*(index);                                  \
            (v0)=(int)ptr[0];                                \
            (v1)=(int)ptr[1];                                \
            (v2)=(int)ptr[2];                                \
            break;                                           \
            }                                                \
          case 8:                                            \
            {                                                \
            const uint64_t *ptr=(const uint64_t *)indices;   \
            ptr+=3*(index);                                  \
            (v0)=(int)ptr[0];                                \
            (v1)=(int)ptr[1];                                \
            (v2)=(int)ptr[2];                                \
            break;                                           \
            }                                                \
          default:                                           \
            {                                                \
            (v0)=(v1)=(v2)=0;                                \
            DROP_DEAD("bad value for sizeOfIndex");          \
            }                                                \
          }

void
Surface_boxRequirements(int *out_indexCount,
                        int *out_vertexCount)
{
*out_indexCount=36;
*out_vertexCount=24;
}

void
Surface_makeBoxGeneric(void *indices,
                       int sizeOfIndex,
                       void *vertices,
                       int sizeOfVertex,
                       void *normals,
                       int sizeOfNormal,
                       void *texCoords,
                       int sizeOfTexCoord,
                       bool counterClockWise,
                       real64_t sizeX,
                       real64_t sizeY,
                       real64_t sizeZ)
{
sizeX*=0.5;
sizeY*=0.5;
sizeZ*=0.5;
if(counterClockWise)
  {
  ADD_V(-sizeX,-sizeY,-sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(0.0,0.0); // Y-
  ADD_V(+sizeX,-sizeY,-sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(1.0,0.0);
  ADD_V(+sizeX,-sizeY,+sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(1.0,1.0);
  ADD_V(-sizeX,-sizeY,+sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(0.0,1.0);
  ADD_TRI(0,1,2); ADD_TRI(0,2,3);
  ADD_V(+sizeX,-sizeY,-sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(0.0,0.0); // X+
  ADD_V(+sizeX,+sizeY,-sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(1.0,0.0);
  ADD_V(+sizeX,+sizeY,+sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(1.0,1.0);
  ADD_V(+sizeX,-sizeY,+sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(0.0,1.0);
  ADD_TRI(4,5,6); ADD_TRI(4,6,7);
  ADD_V(+sizeX,+sizeY,-sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(0.0,0.0); // Y+
  ADD_V(-sizeX,+sizeY,-sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(1.0,0.0);
  ADD_V(-sizeX,+sizeY,+sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(1.0,1.0);
  ADD_V(+sizeX,+sizeY,+sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(0.0,1.0);
  ADD_TRI(8,9,10); ADD_TRI(8,10,11);
  ADD_V(-sizeX,+sizeY,-sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(0.0,0.0); // X-
  ADD_V(-sizeX,-sizeY,-sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(1.0,0.0);
  ADD_V(-sizeX,-sizeY,+sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(1.0,1.0);
  ADD_V(-sizeX,+sizeY,+sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(0.0,1.0);
  ADD_TRI(12,13,14); ADD_TRI(12,14,15);
  ADD_V(+sizeX,-sizeY,+sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(0.0,0.0); // Z+
  ADD_V(+sizeX,+sizeY,+sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(1.0,0.0);
  ADD_V(-sizeX,+sizeY,+sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(1.0,1.0);
  ADD_V(-sizeX,-sizeY,+sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(0.0,1.0);
  ADD_TRI(16,17,18); ADD_TRI(16,18,19);
  ADD_V(-sizeX,-sizeY,-sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(0.0,0.0); // Z-
  ADD_V(-sizeX,+sizeY,-sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(1.0,0.0);
  ADD_V(+sizeX,+sizeY,-sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(1.0,1.0);
  ADD_V(+sizeX,-sizeY,-sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(0.0,1.0);
  ADD_TRI(20,21,22); ADD_TRI(20,22,23);
  }
else
  {
  ADD_V(-sizeX,-sizeY,-sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(1.0,0.0); // Y-
  ADD_V(+sizeX,-sizeY,-sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(0.0,0.0);
  ADD_V(+sizeX,-sizeY,+sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(0.0,1.0);
  ADD_V(-sizeX,-sizeY,+sizeZ); ADD_N(+0.0,+1.0,+0.0); ADD_T(1.0,1.0);
  ADD_TRI(1,0,3); ADD_TRI(1,3,2);
  ADD_V(+sizeX,-sizeY,-sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(1.0,0.0); // X+
  ADD_V(+sizeX,+sizeY,-sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(0.0,0.0);
  ADD_V(+sizeX,+sizeY,+sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(0.0,1.0);
  ADD_V(+sizeX,-sizeY,+sizeZ); ADD_N(-1.0,+0.0,+0.0); ADD_T(1.0,1.0);
  ADD_TRI(5,4,7); ADD_TRI(5,7,6);
  ADD_V(+sizeX,+sizeY,-sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(1.0,0.0); // Y+
  ADD_V(-sizeX,+sizeY,-sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(0.0,0.0);
  ADD_V(-sizeX,+sizeY,+sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(0.0,1.0);
  ADD_V(+sizeX,+sizeY,+sizeZ); ADD_N(+0.0,-1.0,+0.0); ADD_T(1.0,1.0);
  ADD_TRI(9,8,11); ADD_TRI(9,11,10);
  ADD_V(-sizeX,+sizeY,-sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(1.0,0.0); // X-
  ADD_V(-sizeX,-sizeY,-sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(0.0,0.0);
  ADD_V(-sizeX,-sizeY,+sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(0.0,1.0);
  ADD_V(-sizeX,+sizeY,+sizeZ); ADD_N(+1.0,+0.0,+0.0); ADD_T(1.0,1.0);
  ADD_TRI(13,12,15); ADD_TRI(13,15,14);
  ADD_V(+sizeX,-sizeY,+sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(0.0,1.0); // Z+
  ADD_V(+sizeX,+sizeY,+sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(1.0,1.0);
  ADD_V(-sizeX,+sizeY,+sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(1.0,0.0);
  ADD_V(-sizeX,-sizeY,+sizeZ); ADD_N(+0.0,+0.0,-1.0); ADD_T(0.0,0.0);
  ADD_TRI(17,16,19); ADD_TRI(17,19,18);
  ADD_V(-sizeX,-sizeY,-sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(0.0,1.0); // Z-
  ADD_V(-sizeX,+sizeY,-sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(1.0,1.0);
  ADD_V(+sizeX,+sizeY,-sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(1.0,0.0);
  ADD_V(+sizeX,-sizeY,-sizeZ); ADD_N(+0.0,+0.0,+1.0); ADD_T(0.0,0.0);
  ADD_TRI(21,20,23); ADD_TRI(21,23,22);
  }
}

void
Surface_allocBoxGeneric(int *out_indexCount,
                        void **indices,
                        int sizeOfIndex,
                        int *out_vertexCount,
                        void **vertices,
                        int sizeOfVertex,
                        void **normals,
                        int sizeOfNormal,
                        void **texCoords,
                        int sizeOfTexCoord,
                        bool counterClockWise,
                        real64_t sizeX,
                        real64_t sizeY,
                        real64_t sizeZ)
{
Surface_boxRequirements(out_indexCount,out_vertexCount);
void *i=indices
        ? *indices=Memory_allocVoid(*out_indexCount*sizeOfIndex)
        : (void *)0;
void *v=vertices
        ? *vertices=Memory_allocVoid(*out_vertexCount*3*sizeOfVertex)
        : (void *)0;
void *n=normals
        ? *normals=Memory_allocVoid(*out_vertexCount*3*sizeOfNormal)
        : (void *)0;
void *t=texCoords
        ? *texCoords=Memory_allocVoid(*out_vertexCount*2*sizeOfTexCoord)
        : (void *)0;
Surface_makeBoxGeneric(i,sizeOfIndex,
                       v,sizeOfVertex,
                       n,sizeOfNormal,
                       t,sizeOfTexCoord,
                       counterClockWise,sizeX,sizeY,sizeZ);
}

void
Surface_sphereRequirements(int *out_indexCount,
                           int *out_vertexCount,
                           int precision)
{
Surface_ellipsoidRequirements(out_indexCount,out_vertexCount,precision);
}

void
Surface_makeSphereGeneric(void *indices,
                          int sizeOfIndex,
                          void *vertices,
                          int sizeOfVertex,
                          void *normals,
                          int sizeOfNormal,
                          void *texCoords,
                          int sizeOfTexCoord,
                          bool counterClockWise,
                          real64_t radius,
                          int precision)
{
Surface_makeEllipsoidGeneric(indices,sizeOfIndex,
                             vertices,sizeOfVertex,
                             normals,sizeOfNormal,
                             texCoords,sizeOfTexCoord,
                             counterClockWise,
                             radius,radius,radius,
                             precision);
}

void
Surface_allocSphereGeneric(int *out_indexCount,
                           void **indices,
                           int sizeOfIndex,
                           int *out_vertexCount,
                           void **vertices,
                           int sizeOfVertex,
                           void **normals,
                           int sizeOfNormal,
                           void **texCoords,
                           int sizeOfTexCoord,
                           bool counterClockWise,
                           real64_t radius,
                           int precision)
{
Surface_allocEllipsoidGeneric(out_indexCount,indices,sizeOfIndex,
                              out_vertexCount,vertices,sizeOfVertex,
                              normals,sizeOfNormal,
                              texCoords,sizeOfTexCoord,
                              counterClockWise,
                              radius,radius,radius,
                              precision);
}

void
Surface_ellipsoidRequirements(int *out_indexCount,
                              int *out_vertexCount,
                              int precision)
{
precision=precision>0 ? 1+(precision-1)/4 : 1;
*out_indexCount=24*precision*precision;
*out_vertexCount=precision*(4*precision+2)+7;
}

void
Surface_makeEllipsoidGeneric(void *indices,
                             int sizeOfIndex,
                             void *vertices,
                             int sizeOfVertex,
                             void *normals,
                             int sizeOfNormal,
                             void *texCoords,
                             int sizeOfTexCoord,
                             bool counterClockWise,
                             real64_t radiusX,
                             real64_t radiusY,
                             real64_t radiusZ,
                             int precision)
{
precision=precision>0 ? 1+(precision-1)/4 : 1;
if(counterClockWise)
  {
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,1.0);  ADD_T(0.125,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,-1.0); ADD_T(0.125,0.0);
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,1.0);  ADD_T(0.375,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,-1.0); ADD_T(0.375,0.0);
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,1.0);  ADD_T(0.625,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,-1.0); ADD_T(0.625,0.0);
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,1.0);  ADD_T(0.875,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,-1.0); ADD_T(0.875,0.0);
  }
else
  {
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,-1.0); ADD_T(0.875,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,1.0);  ADD_T(0.875,0.0);
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,-1.0); ADD_T(0.625,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,1.0);  ADD_T(0.625,0.0);
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,-1.0); ADD_T(0.375,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,1.0);  ADD_T(0.375,0.0);
  ADD_V(0.0,0.0,radiusZ);  ADD_N(0.0,0.0,-1.0); ADD_T(0.125,1.0);
  ADD_V(0.0,0.0,-radiusZ); ADD_N(0.0,0.0,1.0);  ADD_T(0.125,0.0);
  }
real64_t nX=radiusY*radiusZ;
real64_t nY=radiusX*radiusZ;
real64_t nZ=radiusX*radiusY;
int vertexCount=8;
for(int i=1;i<=precision;++i)
  {
  real64_t t1=i/(real64_t)precision;
  real64_t s1,c1;
  sincos64(M_PI_2*t1,&s1,&c1);
  for(int j=0;j<=4*i;++j)
    {
    real64_t t2=j/(real64_t)(4*i);
    real64_t s2,c2;
    sincos64(M_2PI*t2,&s2,&c2);
    ++vertexCount;
    ADD_V(-s1*c2*radiusX,-s1*s2*radiusY,c1*radiusZ);
    if(counterClockWise)
      {
      ADD_N(-s1*c2*nX,-s1*s2*nY,c1*nZ);
      ADD_T(t2,1.0-0.5*t1);
      }
    else
      {
      ADD_N(s1*c2*nX,s1*s2*nY,-c1*nZ);
      ADD_T(1.0-t2,1.0-0.5*t1);
      }
    if(i!=precision)
      {
      ++vertexCount;
      ADD_V(-s1*c2*radiusX,-s1*s2*radiusY,-c1*radiusZ);
      if(counterClockWise)
        {
        ADD_N(-s1*c2*nX,-s1*s2*nY,-c1*nZ);
        ADD_T(t2,0.5*t1);
        }
      else
        {
        ADD_N(s1*c2*nX,s1*s2*nY,c1*nZ);
        ADD_T(1.0-t2,0.5*t1);
        }
      }
    }
  }
int step=1;
int curLine1=vertexCount-4*precision-1;
int curLine2=curLine1;
for(int i=precision;i--;)
  {
  int nextLine1=i ? curLine1-8*i-2 : 0;
  int nextLine2=nextLine1+1;
  int curPos1=curLine1;
  int curPos2=curLine2;
  int nextPos1=nextLine1;
  int nextPos2=nextLine2;
  for(int k=0;k<4;++k)
    {
    if(counterClockWise)
      {
      ADD_TRI(nextPos1,curPos1,curPos1+step);
      ADD_TRI(curPos2,nextPos2,curPos2+step);
      }
    else
      {
      ADD_TRI(curPos1,nextPos1,curPos1+step);
      ADD_TRI(nextPos2,curPos2,curPos2+step);
      }
    curPos1+=step;
    curPos2+=step;
    for(int j=0;j<i;++j)
      {
      if(counterClockWise)
        {
        ADD_TRI(nextPos1,curPos1,nextPos1+2);
        ADD_TRI(nextPos1+2,curPos1,curPos1+step);
        ADD_TRI(curPos2,nextPos2,nextPos2+2);
        ADD_TRI(curPos2,nextPos2+2,curPos2+step);
        }
      else
        {
        ADD_TRI(curPos1,nextPos1,nextPos1+2);
        ADD_TRI(curPos1,nextPos1+2,curPos1+step);
        ADD_TRI(nextPos2,curPos2,nextPos2+2);
        ADD_TRI(nextPos2+2,curPos2,curPos2+step);
        }
      curPos1+=step;
      curPos2+=step;
      nextPos1+=2;
      nextPos2+=2;
      }
    }
  curLine1=nextLine1;
  curLine2=nextLine2;
  step=2;
  }
}

void
Surface_allocEllipsoidGeneric(int *out_indexCount,
                              void **indices,
                              int sizeOfIndex,
                              int *out_vertexCount,
                              void **vertices,
                              int sizeOfVertex,
                              void **normals,
                              int sizeOfNormal,
                              void **texCoords,
                              int sizeOfTexCoord,
                              bool counterClockWise,
                              real64_t radiusX,
                              real64_t radiusY,
                              real64_t radiusZ,
                              int precision)
{
Surface_ellipsoidRequirements(out_indexCount,out_vertexCount,precision);
void *i=indices
        ? *indices=Memory_allocVoid(*out_indexCount*sizeOfIndex)
        : (void *)0;
void *v=vertices
        ? *vertices=Memory_allocVoid(*out_vertexCount*3*sizeOfVertex)
        : (void *)0;
void *n=normals
        ? *normals=Memory_allocVoid(*out_vertexCount*3*sizeOfNormal)
        : (void *)0;
void *t=texCoords
        ? *texCoords=Memory_allocVoid(*out_vertexCount*2*sizeOfTexCoord)
        : (void *)0;
Surface_makeEllipsoidGeneric(i,sizeOfIndex,
                             v,sizeOfVertex,
                             n,sizeOfNormal,
                             t,sizeOfTexCoord,
                             counterClockWise,
                             radiusX,radiusY,radiusZ,
                             precision);
}

void
Surface_coneRequirements(int *out_indexCount,
                         int *out_vertexCount,
                         bool side,
                         bool bottom,
                         int precision)
{
precision=precision<3 ? 3 : precision;
*out_indexCount=0;
*out_vertexCount=0;
if(side)
  {
  *out_indexCount+=3*precision;
  *out_vertexCount+=2*precision+1;
  }
if(bottom)
  {
  *out_indexCount+=3*precision;
  *out_vertexCount+=precision+2;
  }
}

void
Surface_makeConeGeneric(void *indices,
                        int sizeOfIndex,
                        void *vertices,
                        int sizeOfVertex,
                        void *normals,
                        int sizeOfNormal,
                        void *texCoords,
                        int sizeOfTexCoord,
                        bool counterClockWise,
                        real64_t radius,
                        real64_t height,
                        bool side,
                        bool bottom,
                        int precision)
{
height*=0.5;
precision=precision<3 ? 3 : precision;
int offset=0;
if(side)
  {
  ADD_V(-radius,0.0,-height);
  if(counterClockWise)
    {
    ADD_N(-1.0,0.0,0.0); ADD_T(0.0,0.0);
    }
  else
    {
    ADD_N(1.0,0.0,0.0); ADD_T(1.0,0.0);
    }
  for(int i=1;i<=precision;++i)
    {
    real64_t t1=i/(real64_t)precision;
    real64_t t2=(i-0.5)/precision;
    real64_t s1,c1,s2,c2;
    sincos64(M_2PI*t1,&s1,&c1);
    sincos64(M_2PI*t2,&s2,&c2);
    ADD_V(0.0,0.0,height);
    ADD_V(-radius*c1,-radius*s1,-height);
    if(counterClockWise)
      {
      ADD_N(-c2,-s2,0.0); ADD_T(t2,1.0);
      ADD_N(-c1,-s1,0.0); ADD_T(t1,0.0);
      ADD_TRI(offset+2*(i-1),offset+2*i,offset+2*i-1);
      }
    else
      {
      ADD_N(c2,s2,0); ADD_T(1.0-t2,1.0);
      ADD_N(c1,s1,0); ADD_T(1.0-t1,0.0);
      ADD_TRI(offset+2*(i-1),offset+2*i-1,offset+2*i);
      }
    }
  offset+=2*precision+1;
  }
if(bottom)
  {
  ADD_V(0.0,0.0,-height);
  ADD_V(-radius,0.0,-height);
  if(counterClockWise)
    {
    ADD_N(0.0,0.0,-1.0); ADD_T(0.5,0.5);
    ADD_N(0.0,0.0,-1.0); ADD_T(0.5,0.0);
    }
  else
    {
    ADD_N(0.0,0.0,1.0); ADD_T(0.5,0.5);
    ADD_N(0.0,0.0,1.0); ADD_T(0.5,1.0);
    }
  for(int i=1;i<=precision;++i)
    {
    real64_t s,c;
    sincos64(i*M_2PI/precision,&s,&c);
    ADD_V(-radius*c,-radius*s,-height);
    if(counterClockWise)
      {
      ADD_N(0.0,0.0,-1.0); ADD_T(0.5*(1.0-s),0.5*(1.0-c));
      ADD_TRI(offset,offset+i+1,offset+i);
      }
    else
      {
      ADD_N(0.0,0.0,1.0); ADD_T(0.5*(1.0-s),0.5*(1.0+c));
      ADD_TRI(offset,offset+i,offset+i+1);
      }
    }
  offset+=precision+2;
  }
}

void
Surface_allocConeGeneric(int *out_indexCount,
                         void **indices,
                         int sizeOfIndex,
                         int *out_vertexCount,
                         void **vertices,
                         int sizeOfVertex,
                         void **normals,
                         int sizeOfNormal,
                         void **texCoords,
                         int sizeOfTexCoord,
                         bool counterClockWise,
                         real64_t radius,
                         real64_t height,
                         bool side,
                         bool bottom,
                         int precision)
{
Surface_coneRequirements(out_indexCount,out_vertexCount,side,bottom,precision);
void *i=indices
        ? *indices=Memory_allocVoid(*out_indexCount*sizeOfIndex)
        : (void *)0;
void *v=vertices
        ? *vertices=Memory_allocVoid(*out_vertexCount*3*sizeOfVertex)
        : (void *)0;
void *n=normals
        ? *normals=Memory_allocVoid(*out_vertexCount*3*sizeOfNormal)
        : (void *)0;
void *t=texCoords
        ? *texCoords=Memory_allocVoid(*out_vertexCount*2*sizeOfTexCoord)
        : (void *)0;
Surface_makeConeGeneric(i,sizeOfIndex,
                        v,sizeOfVertex,
                        n,sizeOfNormal,
                        t,sizeOfTexCoord,
                        counterClockWise,radius,height,side,bottom,precision);
}

void
Surface_cylinderRequirements(int *out_indexCount,
                             int *out_vertexCount,
                             bool side,
                             bool top,
                             bool bottom,
                             int precision)
{
precision=precision<3 ? 3 : precision;
*out_indexCount=0;
*out_vertexCount=0;
if(side)
  {
  *out_indexCount+=6*precision;
  *out_vertexCount+=2*precision+2;
  }
if(top)
  {
  *out_indexCount+=3*precision;
  *out_vertexCount+=precision+2;
  }
if(bottom)
  {
  *out_indexCount+=3*precision;
  *out_vertexCount+=precision+2;
  }
}

void
Surface_makeCylinderGeneric(void *indices,
                            int sizeOfIndex,
                            void *vertices,
                            int sizeOfVertex,
                            void *normals,
                            int sizeOfNormal,
                            void *texCoords,
                            int sizeOfTexCoord,
                            bool counterClockWise,
                            real64_t radius,
                            real64_t height,
                            bool side,
                            bool top,
                            bool bottom,
                            int precision)
{
height*=0.5;
precision=precision<3 ? 3 : precision;
int offset=0;
if(side)
  {
  ADD_V(-radius,0.0,-height);
  ADD_V(-radius,0.0,height);
  if(counterClockWise)
    {
    ADD_N(-1.0,0.0,0.0); ADD_T(0.0,0.0);
    ADD_N(-1.0,0.0,0.0); ADD_T(0.0,1.0);
    }
  else
    {
    ADD_N(1.0,0.0,0.0); ADD_T(1.0,0.0);
    ADD_N(1.0,0.0,0.0); ADD_T(1.0,1.0);
    }
  for(int i=1;i<=precision;++i)
    {
    real64_t t=i/(real64_t)precision;
    real64_t s,c;
    sincos64(M_2PI*t,&s,&c);
    ADD_V(-radius*c,-radius*s,-height);
    ADD_V(-radius*c,-radius*s,height);
    if(counterClockWise)
      {
      ADD_N(-c,-s,0.0); ADD_T(t,0.0);
      ADD_N(-c,-s,0.0); ADD_T(t,1.0);
      ADD_TRI(offset+2*(i-1),offset+2*i,offset+2*i+1);
      ADD_TRI(offset+2*(i-1),offset+2*i+1,offset+2*(i-1)+1);
      }
    else
      {
      ADD_N(c,s,0); ADD_T(1.0-t,0.0);
      ADD_N(c,s,0); ADD_T(1.0-t,1.0);
      ADD_TRI(offset+2*i,offset+2*(i-1),offset+2*(i-1)+1);
      ADD_TRI(offset+2*i,offset+2*(i-1)+1,offset+2*i+1);
      }
    }
  offset+=2*precision+2;
  }
if(top)
  {
  ADD_V(0.0,0.0,height);
  ADD_V(-radius,0.0,height);
  if(counterClockWise)
    {
    ADD_N(0.0,0.0,1.0); ADD_T(0.5,0.5);
    ADD_N(0.0,0.0,1.0); ADD_T(0.5,1.0);
    }
  else
    {
    ADD_N(0.0,0.0,-1.0); ADD_T(0.5,0.5);
    ADD_N(0.0,0.0,-1.0); ADD_T(0.5,0.0);
    }
  for(int i=1;i<=precision;++i)
    {
    real64_t s,c;
    sincos64(i*M_2PI/precision,&s,&c);
    ADD_V(-radius*c,-radius*s,height);
    if(counterClockWise)
      {
      ADD_N(0.0,0.0,1.0); ADD_T(0.5*(1.0-s),0.5*(1.0+c));
      ADD_TRI(offset,offset+i,offset+i+1);
      }
    else
      {
      ADD_N(0.0,0.0,-1.0); ADD_T(0.5*(1.0-s),0.5*(1.0-c));
      ADD_TRI(offset,offset+i+1,offset+i);
      }
    }
  offset+=precision+2;
  }
if(bottom)
  {
  ADD_V(0.0,0.0,-height);
  ADD_V(-radius,0.0,-height);
  if(counterClockWise)
    {
    ADD_N(0.0,0.0,-1.0); ADD_T(0.5,0.5);
    ADD_N(0.0,0.0,-1.0); ADD_T(0.5,0.0);
    }
  else
    {
    ADD_N(0.0,0.0,1.0); ADD_T(0.5,0.5);
    ADD_N(0.0,0.0,1.0); ADD_T(0.5,1.0);
    }
  for(int i=1;i<=precision;++i)
    {
    real64_t s,c;
    sincos64(i*M_2PI/precision,&s,&c);
    ADD_V(-radius*c,-radius*s,-height);
    if(counterClockWise)
      {
      ADD_N(0.0,0.0,-1.0); ADD_T(0.5*(1.0-s),0.5*(1.0-c));
      ADD_TRI(offset,offset+i+1,offset+i);
      }
    else
      {
      ADD_N(0.0,0.0,1.0); ADD_T(0.5*(1.0-s),0.5*(1.0+c));
      ADD_TRI(offset,offset+i,offset+i+1);
      }
    }
  offset+=precision+2;
  }
}

void
Surface_allocCylinderGeneric(int *out_indexCount,
                             void **indices,
                             int sizeOfIndex,
                             int *out_vertexCount,
                             void **vertices,
                             int sizeOfVertex,
                             void **normals,
                             int sizeOfNormal,
                             void **texCoords,
                             int sizeOfTexCoord,
                             bool counterClockWise,
                             real64_t radius,
                             real64_t height,
                             bool side,
                             bool top,
                             bool bottom,
                             int precision)
{
Surface_cylinderRequirements(out_indexCount,out_vertexCount,
                             side,top,bottom,precision);
void *i=indices
        ? *indices=Memory_allocVoid(*out_indexCount*sizeOfIndex)
        : (void *)0;
void *v=vertices
        ? *vertices=Memory_allocVoid(*out_vertexCount*3*sizeOfVertex)
        : (void *)0;
void *n=normals
        ? *normals=Memory_allocVoid(*out_vertexCount*3*sizeOfNormal)
        : (void *)0;
void *t=texCoords
        ? *texCoords=Memory_allocVoid(*out_vertexCount*2*sizeOfTexCoord)
        : (void *)0;
Surface_makeCylinderGeneric(i,sizeOfIndex,
                            v,sizeOfVertex,
                            n,sizeOfNormal,
                            t,sizeOfTexCoord,
                            counterClockWise,radius,height,
                            side,top,bottom,precision);
}

void
Surface_torusRequirements(int *out_indexCount,
                          int *out_vertexCount,
                          int precision1,
                          int precision2)
{
precision1=precision1<3 ? 3 : precision1;
precision2=precision2<3 ? 3 : precision2;
*out_indexCount=6*precision1*precision2;
*out_vertexCount=(precision1+1)*(precision2+1);
}

void
Surface_makeTorusGeneric(void *indices,
                         int sizeOfIndex,
                         void *vertices,
                         int sizeOfVertex,
                         void *normals,
                         int sizeOfNormal,
                         void *texCoords,
                         int sizeOfTexCoord,
                         bool counterClockWise,
                         real64_t radius1,
                         real64_t radius2,
                         int precision1,
                         int precision2)
{
precision1=precision1<3 ? 3 : precision1;
precision2=precision2<3 ? 3 : precision2;
for(int i1=0;i1<=precision1;++i1)
  {
  real64_t t1=i1/(real64_t)precision1;
  real64_t s1,c1;
  sincos64(M_2PI*t1,&s1,&c1);
  for(int i2=0;i2<=precision2;++i2)
    {
    real64_t t2=i2/(real64_t)precision2;
    real64_t s2,c2;
    sincos64(M_2PI*t2,&s2,&c2);
    ADD_V(c1*(radius2*c2-radius1),s1*(radius2*c2-radius1),radius2*s2);
    if(counterClockWise)
      {
      ADD_N(c2*c1,c2*s1,s2);
      ADD_T(t1,1.0-t2);
      if(i1&&i2)
        {
        ADD_TRI((i1-1)*(precision2+1)+(i2-1),
                (i1-1)*(precision2+1)+i2,
                i1*(precision2+1)+i2);
        ADD_TRI((i1-1)*(precision2+1)+(i2-1),
                i1*(precision2+1)+i2,
                i1*(precision2+1)+(i2-1));
        }
      }
    else
      {
      ADD_N(-c2*c1,-c2*s1,-s2);
      ADD_T(1.0-t1,1.0-t2);
      if(i1&&i2)
        {
        ADD_TRI((i1-1)*(precision2+1)+(i2-1),
                i1*(precision2+1)+i2,
                (i1-1)*(precision2+1)+i2);
        ADD_TRI((i1-1)*(precision2+1)+(i2-1),
                i1*(precision2+1)+(i2-1),
                i1*(precision2+1)+i2);
        }
      }
    }
  }
}

void
Surface_allocTorusGeneric(int *out_indexCount,
                          void **indices,
                          int sizeOfIndex,
                          int *out_vertexCount,
                          void **vertices,
                          int sizeOfVertex,
                          void **normals,
                          int sizeOfNormal,
                          void **texCoords,
                          int sizeOfTexCoord,
                          bool counterClockWise,
                          real64_t radius1,
                          real64_t radius2,
                          int precision1,
                          int precision2)
{
Surface_torusRequirements(out_indexCount,out_vertexCount,
                          precision1,precision2);
void *i=indices
        ? *indices=Memory_allocVoid(*out_indexCount*sizeOfIndex)
        : (void *)0;
void *v=vertices
        ? *vertices=Memory_allocVoid(*out_vertexCount*3*sizeOfVertex)
        : (void *)0;
void *n=normals
        ? *normals=Memory_allocVoid(*out_vertexCount*3*sizeOfNormal)
        : (void *)0;
void *t=texCoords
        ? *texCoords=Memory_allocVoid(*out_vertexCount*2*sizeOfTexCoord)
        : (void *)0;
Surface_makeTorusGeneric(i,sizeOfIndex,
                         v,sizeOfVertex,
                         n,sizeOfNormal,
                         t,sizeOfTexCoord,
                         counterClockWise,radius1,radius2,
                         precision1,precision2);
}

void
Surface_gridRequirements(int *out_indexCount,
                         int *out_vertexCount,
                         int precisionX,
                         int precisionY)
{
precisionX=precisionX>0 ? precisionX : 1;
precisionY=precisionY>0 ? precisionY : 1;
*out_indexCount=6*precisionX*precisionY;
*out_vertexCount=(precisionX+1)*(precisionY+1);
}

void
Surface_makeGridGeneric(void *indices,
                        int sizeOfIndex,
                        void *vertices,
                        int sizeOfVertex,
                        void *normals,
                        int sizeOfNormal,
                        void *texCoords,
                        int sizeOfTexCoord,
                        bool counterClockWise,
                        real64_t sizeX,
                        real64_t sizeY,
                        int precisionX,
                        int precisionY)
{
precisionX=precisionX>0 ? precisionX : 1;
precisionY=precisionY>0 ? precisionY : 1;
int line=precisionX+1;
real64_t xMin=-0.5*sizeX,xStep=1.0/precisionX,xStep2=sizeX*xStep;
real64_t yMin=-0.5*sizeY,yStep=1.0/precisionY,yStep2=sizeY*yStep;
for(int y=0;y<=precisionY;++y)
  {
  for(int x=0;x<=precisionX;++x)
    {
    ADD_V(xMin+x*xStep2,yMin+y*yStep2,0.0)
    if(counterClockWise)
      {
      ADD_N(0.0,0.0,1.0);
      ADD_T(x*xStep,y*yStep);
      }
    else
      {
      ADD_N(0.0,0.0,-1.0);
      ADD_T(1.0-(x*xStep),y*yStep);
      }
    if(x&&y)
      {
      int p0=(y-1)*line+(x-1);
      int p1=(y-1)*line+x;
      int p2=y*line+x;
      int p3=y*line+(x-1);
      if(counterClockWise)
        {
        ADD_TRI(p0,p1,p2)
        ADD_TRI(p0,p2,p3)
        }
      else
        {
        ADD_TRI(p0,p2,p1)
        ADD_TRI(p0,p3,p2)
        }
      }
    }
  }
}

void
Surface_allocGridGeneric(int *out_indexCount,
                         void **indices,
                         int sizeOfIndex,
                         int *out_vertexCount,
                         void **vertices,
                         int sizeOfVertex,
                         void **normals,
                         int sizeOfNormal,
                         void **texCoords,
                         int sizeOfTexCoord,
                         bool counterClockWise,
                         real64_t sizeX,
                         real64_t sizeY,
                         int precisionX,
                         int precisionY)
{
Surface_gridRequirements(out_indexCount,out_vertexCount,
                         precisionX,precisionY);
void *i=indices
        ? *indices=Memory_allocVoid(*out_indexCount*sizeOfIndex)
        : (void *)0;
void *v=vertices
        ? *vertices=Memory_allocVoid(*out_vertexCount*3*sizeOfVertex)
        : (void *)0;
void *n=normals
        ? *normals=Memory_allocVoid(*out_vertexCount*3*sizeOfNormal)
        : (void *)0;
void *t=texCoords
        ? *texCoords=Memory_allocVoid(*out_vertexCount*2*sizeOfTexCoord)
        : (void *)0;
Surface_makeGridGeneric(i,sizeOfIndex,
                        v,sizeOfVertex,
                        n,sizeOfNormal,
                        t,sizeOfTexCoord,
                        counterClockWise,sizeX,sizeY,precisionX,precisionY);
}

void
Surface_loadMeshGeneric(const char *fileName,
                        int *out_indexCount,
                        void **out_indices,
                        int sizeOfIndex,
                        int *out_vertexCount,
                        void **out_vertices,
                        int sizeOfVertex,
                        void **out_normals,
                        int sizeOfNormal,
                        void **out_texCoords,
                        int sizeOfTexCoord)
{
*out_indexCount=0;
void *indices=(void *)0;
if(out_indices)
  {
  *out_indices=indices;
  }
*out_vertexCount=0;
void *vertices=(void *)0;
if(out_vertices)
  {
  *out_vertices=vertices;
  }
void *normals=(void *)0;
if(out_normals)
  {
  *out_normals=normals;
  }
void *texCoords=(void *)0;
if(out_texCoords)
  {
  *out_texCoords=texCoords;
  }
FILE *f=Path_fopen(fileName,"r");
if(!f)
  { SYS_DROP_DEAD("Path_fopen() failure"); }
bool hasNormals=false;
bool hasTexCoords=false;
int nbV=0,vCount=0;
int nbT=0,tCount=0;
for(;;)
  {
  char line[0x100];
  if(!fgets(line,0x100,f)) break;
  const char *pos=strchr(line,'#');
  if(pos)
    {
    line[pos-line]='\0'; // cut line to ignore comment
    }
  if(line[0]&&strcmp(line,"\n")&&strcmp(line,"\r\n")) // ignore empty lines
    {
    if(!nbV||!nbT)
      {
      int charCount=0;
      if((sscanf(line," %d triangles %d vertices %n",
                 &nbT,&nbV,&charCount)!=2)||!charCount)
        { DROP_DEAD("triangle and vertex counts expected"); }
      if(out_normals)
        {
        if(strstr(line,"normals"))
          {
          hasNormals=true;
          }
        // else normals will be computed
        }
      if(out_texCoords)
        {
        if(strstr(line,"texCoords"))
          {
          hasTexCoords=true;
          }
        else
          { DROP_DEAD("texCoords expected"); }
        }
      *out_vertexCount=nbV;
      if(out_vertices)
        {
        *out_vertices=vertices=Memory_allocVoid(3*nbV*sizeOfVertex);
        }
      if(out_normals)
        {
        *out_normals=normals=Memory_allocVoid(3*nbV*sizeOfNormal);
        }
      if(out_texCoords)
        {
        *out_texCoords=texCoords=Memory_allocVoid(2*nbV*sizeOfTexCoord);
        }
      *out_indexCount=3*nbT;
      if(out_indices)
        {
        *out_indices=indices=Memory_allocVoid(3*nbT*sizeOfIndex);
        }
      }
    else if(vCount<nbV)
      {
      real64_t vx,vy,vz;
      const char *pos=strchr(line,'v');
      if(!pos||(sscanf(line," v %lg %lg %lg ",&vx,&vy,&vz)!=3))
        { DROP_DEAD("vertex expected"); }
      ADD_V(vx,vy,vz)
      if(hasNormals)
        {
        real64_t nx,ny,nz;
        pos=strchr(line,'n');
        if(!pos||(sscanf(pos," n %lg %lg %lg ",&nx,&ny,&nz)!=3))
          { DROP_DEAD("normal expected"); }
        ADD_N(nx,ny,nz)
        }
      if(hasTexCoords)
        {
        real64_t tx,ty;
        pos=strchr(line,'t');
        if(!pos||(sscanf(pos," t %lg %lg ",&tx,&ty)!=2))
          { DROP_DEAD("texCoord expected"); }
        ADD_T(tx,ty)
        }
      ++vCount;
      }
    else if(tCount<nbT)
      {
      int i0,i1,i2;
      if(sscanf(line," i %d %d %d ",&i0,&i1,&i2)!=3)
        { DROP_DEAD("triangle expected"); }
      ADD_TRI(i0,i1,i2)
      ++tCount;
      }
    else
      {
      DROP_DEAD("unexpected line");
      }
    }
  }
fclose(f);
if(!nbV||!nbT||(vCount<nbV)||(tCount<nbT))
  { DROP_DEAD("incomplete mesh"); }
if(normals&&!hasNormals)
  {
  Surface_computeDefaultNormalsGeneric(*out_indexCount,
                                       *out_indices,sizeOfIndex,
                                       *out_vertexCount,
                                       *out_vertices,sizeOfVertex,
                                       *out_normals,sizeOfNormal);
  }
}

void
Surface_computeDefaultNormalsGeneric(int indexCount,
                                     const void *indices,
                                     int sizeOfIndex,
                                     int vertexCount,
                                     const void *vertices,
                                     int sizeOfVertex,
                                     void *normals,
                                     int sizeOfNormal)
{
int i;
int triangleCount=indexCount/3;
real64_t *n=(sizeOfNormal==8)
           ? (real64_t *)normals
           : Memory_alloc(vertexCount*3,real64_t); // keep precision
memset(n,0,vertexCount*3*sizeof(real64_t));
for(i=0;i<triangleCount;++i) // sum normals per vertex
  {
  int p0,p1,p2;
  GET_TRI(p0,p1,p2,i)
  real64_t x0,y0,z0,x1,y1,z1,x2,y2,z2;
  GET_V(x0,y0,z0,p0)
  GET_V(x1,y1,z1,p1)
  GET_V(x2,y2,z2,p2)
  real64_t dx1=x1-x0, dy1=y1-y0, dz1=z1-z0;
  real64_t dx2=x2-x0, dy2=y2-y0, dz2=z2-z0;
  real64_t dx0=dy1*dz2-dz1*dy2; // normal per triangle
  real64_t dy0=dx2*dz1-dz2*dx1;
  real64_t dz0=dx1*dy2-dy1*dx2;
  // don't normalise now to take triangle surface into account
#if 0
  int j=3*p0;
  n[j+0]+=dx0;
  n[j+1]+=dy0;
  n[j+2]+=dz0;
  j=3*p1;
  n[j+0]+=dx0;
  n[j+1]+=dy0;
  n[j+2]+=dz0;
  j=3*p2;
  n[j+0]+=dx0;
  n[j+1]+=dy0;
  n[j+2]+=dz0;
#else
  for(int j=0;j<vertexCount;++j)
    {
    real64_t x,y,z;
    GET_V(x,y,z,j)
    if(((x==x0)&&(y==y0)&&(z==z0))||
       ((x==x1)&&(y==y1)&&(z==z1))||
       ((x==x2)&&(y==y2)&&(z==z2)))
      {
      n[3*j+0]+=dx0;
      n[3*j+1]+=dy0;
      n[3*j+2]+=dz0;
      }
    }
#endif
  }
for(i=0;i<vertexCount;++i) // normalise
  {
  int j=3*i;
  real64_t nx=n[j+0];
  real64_t ny=n[j+1];
  real64_t nz=n[j+2];
  real64_t r=sqrt(nx*nx+ny*ny+nz*nz);
  r=r ? 1.0/r : 1.0;
  ADD_N(nx*r,ny*r,nz*r);
  }
if(n!=normals)
  {
  Memory_free(n);
  }
}

//----------------------------------------------------------------------------
